# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from typing import Any, Union


class Cache(object):
    """Simple caching mechanism to speed-up experiments by re-using evaluated configurations
    and storing atomic results in separate files.

    Example:
    ```
    # `x` - (list) one-hot encoded configuration
    cache_key = Cache.key(x)

    if not self.cache.exists(cache_key):
        top1, top5, gflops, model_params = self.evaluator.validate_subnet(subnet_sample)
        latency = self.evaluator.benchmark_subnet()
        date = str(datetime.now())
        self.cache.update(key=cache_key, payload=[subnet_sample, date, float(latency), float(top1)])
    else:
        subnet_sample, date, latency, top1 = self.cache.get(key=cache_key)
    ```
    """

    _hits = 0

    def __init__(self, name: str, cache_dir: str = '/store/.torch/dynast_cache') -> None:
        """
        Params:
        - `name` - (str) unique name of your experiment. Cache name should be unique
            for each platform and experiment configuration.
        - `cache_dir` - (str, optional) root directory where cache will be stored.
        """
        self.name = name
        self.cache_dir = cache_dir
        os.makedirs(os.path.join(self.cache_dir, self.name), exist_ok=True)

    def update(self, key: str, payload: Any) -> None:
        """Update (add new or replace existing) cache stored under `key`.

        Params:
        - `key` - (str) key under which `payload` will be stored
        - `payload` - (Any) object to be cached. The only requirement is for `payload`
            to be representable in the JSON format.
        """
        with open(self._get_key_path(key), 'w') as f:
            return json.dump(payload, f, indent=4)

    def get(self, key: str) -> Union[Any, None]:
        """Retrieve cache stored under `key`. Returns `None` if 'key` doesnt exist.

        Params:
        - `key` - (str) `key` to retrieve payload.
        """
        try:
            with open(self._get_key_path(key), 'r') as f:
                self._hits += 1
                return json.load(f)
        except FileNotFoundError:
            return None

    @staticmethod
    def key(o: list) -> str:
        """Translate PyMoo vector of one-hot encoded configuration into a file-name friendly string.

        Params:
        `o` - (list) Network's one-hot configuration
        """
        return '_'.join([str(_) for _ in o])

    def exists(self, key: str) -> bool:
        """Checks if cache `key` exists.

        Params:
        - `key` - (str) `key` to retrieve payload.
        """
        return os.path.exists(self._get_key_path(key))

    def _get_key_path(self, key: str) -> str:
        """
        Params:
        - `key` - (str) `key` under which the payload is/will be stored.
        """
        return os.path.join(self.cache_dir, self.name, '{}.json'.format(key))



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import ast

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

from dynast.search.encoding import EncodingBase
from dynast.utils import log


class BertSST2Encoding(EncodingBase):
    def __init__(self, param_dict: dict, verbose: bool = False, seed: int = 0):
        super().__init__(param_dict, verbose, seed)

    def onehot_custom(self, subnet_cfg, provide_onehot=True, max_layers=12):
        features = []
        num_layers = subnet_cfg['num_layers'][0]

        attn_head_list = subnet_cfg['num_attention_heads'][:num_layers] + [0] * (max_layers - num_layers)
        intermediate_size_list = subnet_cfg['intermediate_size'][:num_layers] + [0] * (max_layers - num_layers)
        features = [num_layers] + attn_head_list + intermediate_size_list

        if provide_onehot == True:
            examples = np.array([features])
            one_hot_count = 0
            # TODO(macsz,sharathns93) `self.unique_values` might be uninitialized
            # if `create_training_set` hasn't been called prior to this
            unique_values = self.unique_values

            for unique in unique_values:
                one_hot_count += len(unique.tolist())

            one_hot_examples = np.zeros((examples.shape[0], one_hot_count))
            for e, example in enumerate(examples):
                offset = 0
                for f in range(len(example)):
                    index = np.where(unique_values[f] == example[f])[0] + offset
                    one_hot_examples[e, index] = 1.0
                    offset += len(unique_values[f])
            return one_hot_examples

        else:
            return features

    def import_csv(self, filepath, config, objective, column_names=None, drop_duplicates=True):
        '''
        Import a csv file generated from a supernetwork search for the purpose
        of training a predictor.

        filepath - path of the csv to be imported.
        config - the subnetwork configuration
        objective - target/label for the subnet configuration (e.g. accuracy, latency)
        column_names - a list of column names for the dataframe
        df - the output dataframe that contains the original config dict, pymoo, and 1-hot
             equivalent vector for training.
        '''

        if column_names == None:
            df = pd.read_csv(filepath)
        else:
            df = pd.read_csv(filepath)
            df.columns = column_names
        df = df[[config, objective]]
        # Old corner case coverage
        df[config] = df[config].replace({'null': 'None'}, regex=True)

        if drop_duplicates:
            df.drop_duplicates(subset=[config], inplace=True)
            df.reset_index(drop=True, inplace=True)

        convert_to_dict = list()
        convert_to_pymoo = list()
        convert_to_onehot = list()
        for i in range(len(df)):
            # Elastic Param Config format
            config_as_dict = ast.literal_eval(df[config].iloc[i])
            convert_to_dict.append(config_as_dict)
            # PyMoo 1-D vector format
            config_as_pymoo = self.translate2pymoo(config_as_dict)
            convert_to_pymoo.append(config_as_pymoo)
            # Onehot predictor format
            config_as_onehot = self.onehot_custom(config_as_dict, provide_onehot=False)
            convert_to_onehot.append(config_as_onehot)
        df[config] = convert_to_dict
        df['config_pymoo'] = convert_to_pymoo
        df['config_onehot'] = convert_to_onehot
        return df

    def create_training_set(self, dataframe, config='subnet', train_with_all=True, split=0.33, seed=None):
        '''
        Create a sklearn compatible test/train set from an imported results csv
        after "import_csv" method is run.
        '''

        collect_rows = list()
        for i in range(len(dataframe)):
            collect_rows.append(np.asarray(dataframe['config_onehot'].iloc[i]))
        features = np.asarray(collect_rows)
        labels = dataframe.drop(columns=[config, 'config_pymoo', 'config_onehot']).values

        assert len(features) == len(labels)
        one_hot_count = 0
        unique_values = []

        for c in range(features.shape[1]):
            unique_values.append(np.unique(features[:, c]))
            one_hot_count += len(unique_values[-1])
        one_hot_examples = np.zeros((features.shape[0], one_hot_count))
        for e, example in enumerate(features):
            offset = 0
            for f in range(len(example)):
                index = np.where(unique_values[f] == example[f])[0] + offset
                one_hot_examples[e, index] = 1.0
                offset += len(unique_values[f])

        features = one_hot_examples
        self.unique_values = unique_values
        if train_with_all:
            log.info('Training set length={}'.format(len(labels)))
            return features, labels
        else:
            features_train, features_test, labels_train, labels_test = train_test_split(
                features, labels, test_size=split, random_state=seed
            )
            log.info('Test ({}) Train ({}) ratio is {}.'.format(len(labels_train), len(labels_test), split))
            return features_train, features_test, labels_train, labels_test



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Once for All: Train One Network and Specialize it for Efficient Deployment
# Han Cai, Chuang Gan, Tianzhe Wang, Zhekai Zhang, Song Han
# International Conference on Learning Representations (ICLR), 2020.

import json

import torch

from dynast.supernetwork.image_classification.ofa.ofa.imagenet_classification.elastic_nn.networks import (
    OFAMobileNetV3,
    OFAProxylessNASNets,
    OFAResNets,
)
from dynast.supernetwork.image_classification.ofa.ofa.imagenet_classification.networks import (
    get_net_by_name,
    proxyless_base,
)
from dynast.supernetwork.image_classification.ofa.ofa.utils import download_url

__all__ = [
    "ofa_specialized",
    "ofa_net",
    "proxylessnas_net",
    "proxylessnas_mobile",
    "proxylessnas_cpu",
    "proxylessnas_gpu",
]


def ofa_specialized(net_id, pretrained=True):
    url_base = "https://hanlab.mit.edu/files/OnceForAll/ofa_specialized/"
    net_config = json.load(
        open(
            download_url(
                url_base + net_id + "/net.config",
                model_dir=".torch/ofa_specialized/%s/" % net_id,
            )
        )
    )
    net = get_net_by_name(net_config["name"]).build_from_config(net_config)

    image_size = json.load(
        open(
            download_url(
                url_base + net_id + "/run.config",
                model_dir=".torch/ofa_specialized/%s/" % net_id,
            )
        )
    )["image_size"]

    if pretrained:
        init = torch.load(
            download_url(
                url_base + net_id + "/init",
                model_dir=".torch/ofa_specialized/%s/" % net_id,
            ),
            map_location="cpu",
        )["state_dict"]
        net.load_state_dict(init)
    return net, image_size


def ofa_net(net_id, pretrained=True):
    if net_id == "ofa_proxyless_d234_e346_k357_w1.3":
        net = OFAProxylessNASNets(
            dropout_rate=0,
            width_mult=1.3,
            ks_list=[3, 5, 7],
            expand_ratio_list=[3, 4, 6],
            depth_list=[2, 3, 4],
        )
    elif net_id == "ofa_mbv3_d234_e346_k357_w1.0":
        net = OFAMobileNetV3(
            dropout_rate=0,
            width_mult=1.0,
            ks_list=[3, 5, 7],
            expand_ratio_list=[3, 4, 6],
            depth_list=[2, 3, 4],
        )
    elif net_id == "ofa_mbv3_d234_e346_k357_w1.2":
        net = OFAMobileNetV3(
            dropout_rate=0,
            width_mult=1.2,
            ks_list=[3, 5, 7],
            expand_ratio_list=[3, 4, 6],
            depth_list=[2, 3, 4],
        )
    elif net_id == "ofa_resnet50":
        net = OFAResNets(
            dropout_rate=0,
            depth_list=[0, 1, 2],
            expand_ratio_list=[0.2, 0.25, 0.35],
            width_mult_list=[0.65, 0.8, 1.0],
        )
        net_id = "ofa_resnet50_d=0+1+2_e=0.2+0.25+0.35_w=0.65+0.8+1.0"
    else:
        raise ValueError("Not supported: %s" % net_id)

    if pretrained:
        url_base = "https://hanlab.mit.edu/files/OnceForAll/ofa_nets/"
        init = torch.load(
            download_url(url_base + net_id, model_dir=".torch/ofa_nets"),
            map_location="cpu",
        )["state_dict"]
        net.load_state_dict(init)
    return net


def proxylessnas_net(net_id, pretrained=True):
    net = proxyless_base(
        net_config="https://hanlab.mit.edu/files/proxylessNAS/%s.config" % net_id,
    )
    if pretrained:
        net.load_state_dict(
            torch.load(
                download_url("https://hanlab.mit.edu/files/proxylessNAS/%s.pth" % net_id),
                map_location="cpu",
            )["state_dict"]
        )
    return net


def proxylessnas_mobile(pretrained=True):
    return proxylessnas_net("proxyless_mobile", pretrained)


def proxylessnas_cpu(pretrained=True):
    return proxylessnas_net("proxyless_cpu", pretrained)


def proxylessnas_gpu(pretrained=True):
    return proxylessnas_net("proxyless_gpu", pretrained)



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
from dynast.analytics.visualize import plot_search_progression

from dynast.utils import log
from datetime import datetime


class EvaluationInterface:
    """
    The interface class update is required to be updated for each unique SuperNetwork
    framework as it controls how evaluation calls are made from DyNAS-T

    Args:
        evaluator : class
            The 'runner' that performs the validation or prediction
        manager : class
            The DyNAS-T manager that translates between PyMoo and the parameter dict
        csv_path : string
            (Optional) The csv file that get written to during the subnetwork search
    """

    def __init__(self, evaluator, manager, optimization_metrics, measurements, csv_path, predictor_mode):
        self.evaluator = evaluator
        self.manager = manager
        self.optimization_metrics = optimization_metrics
        self.measurements = measurements
        self.predictor_mode = predictor_mode
        self.csv_path = csv_path

    def format_csv(self, csv_header):
        if self.csv_path:
            f = open(self.csv_path, "w")
            writer = csv.writer(f)
            result = csv_header
            writer.writerow(result)
            f.close()
        log.info(f'(Re)Formatted results file: {self.csv_path}')
        log.info(f'csv file header: {csv_header}')

    def write_results(self, subnet_sample, supernet_metrics, individual_results, plot: bool = True):
        if not self.csv_path:
            return

        with open(self.csv_path, 'a') as f:
            writer = csv.writer(f)
            date = str(datetime.now())
            result = [
                subnet_sample,
                date,
            ]

            for suppernet_metric in supernet_metrics[self.manager.supernet]:
                result += [individual_results[suppernet_metric]]
            writer.writerow(result)
            log.info(f'Results updated: {self.csv_path}')

        if plot:
            plot_search_progression(

            )



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Once for All: Train One Network and Specialize it for Efficient Deployment
# Han Cai, Chuang Gan, Tianzhe Wang, Zhekai Zhang, Song Han
# International Conference on Learning Representations (ICLR), 2020.

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter

from dynast.supernetwork.image_classification.ofa.ofa.utils import (
    MyConv2d,
    MyNetwork,
    SEModule,
    get_same_padding,
    make_divisible,
    sub_filter_start_end,
)

__all__ = [
    "DynamicSeparableConv2d",
    "DynamicConv2d",
    "DynamicGroupConv2d",
    "DynamicBatchNorm2d",
    "DynamicGroupNorm",
    "DynamicSE",
    "DynamicLinear",
]


class DynamicSeparableConv2d(nn.Module):
    KERNEL_TRANSFORM_MODE = 1  # None or 1

    def __init__(self, max_in_channels, kernel_size_list, stride=1, dilation=1):
        super(DynamicSeparableConv2d, self).__init__()

        self.max_in_channels = max_in_channels
        self.kernel_size_list = kernel_size_list
        self.stride = stride
        self.dilation = dilation

        self.conv = nn.Conv2d(
            self.max_in_channels,
            self.max_in_channels,
            max(self.kernel_size_list),
            self.stride,
            groups=self.max_in_channels,
            bias=False,
        )

        self._ks_set = list(set(self.kernel_size_list))
        self._ks_set.sort()  # e.g., [3, 5, 7]
        if self.KERNEL_TRANSFORM_MODE is not None:
            # register scaling parameters
            # 7to5_matrix, 5to3_matrix
            scale_params = {}
            for i in range(len(self._ks_set) - 1):
                ks_small = self._ks_set[i]
                ks_larger = self._ks_set[i + 1]
                param_name = "%dto%d" % (ks_larger, ks_small)
                # noinspection PyArgumentList
                scale_params["%s_matrix" % param_name] = Parameter(torch.eye(ks_small**2))
            for name, param in scale_params.items():
                self.register_parameter(name, param)

        self.active_kernel_size = max(self.kernel_size_list)

    def get_active_filter(self, in_channel, kernel_size):
        out_channel = in_channel
        max_kernel_size = max(self.kernel_size_list)

        start, end = sub_filter_start_end(max_kernel_size, kernel_size)
        filters = self.conv.weight[:out_channel, :in_channel, start:end, start:end]
        if self.KERNEL_TRANSFORM_MODE is not None and kernel_size < max_kernel_size:
            start_filter = self.conv.weight[:out_channel, :in_channel, :, :]  # start with max kernel
            for i in range(len(self._ks_set) - 1, 0, -1):
                src_ks = self._ks_set[i]
                if src_ks <= kernel_size:
                    break
                target_ks = self._ks_set[i - 1]
                start, end = sub_filter_start_end(src_ks, target_ks)
                _input_filter = start_filter[:, :, start:end, start:end]
                _input_filter = _input_filter.contiguous()
                _input_filter = _input_filter.view(_input_filter.size(0), _input_filter.size(1), -1)
                _input_filter = _input_filter.view(-1, _input_filter.size(2))
                _input_filter = F.linear(
                    _input_filter,
                    self.__getattr__("%dto%d_matrix" % (src_ks, target_ks)),
                )
                _input_filter = _input_filter.view(filters.size(0), filters.size(1), target_ks**2)
                _input_filter = _input_filter.view(filters.size(0), filters.size(1), target_ks, target_ks)
                start_filter = _input_filter
            filters = start_filter
        return filters

    def forward(self, x, kernel_size=None):
        if kernel_size is None:
            kernel_size = self.active_kernel_size
        in_channel = x.size(1)

        filters = self.get_active_filter(in_channel, kernel_size).contiguous()

        padding = get_same_padding(kernel_size)
        filters = self.conv.weight_standardization(filters) if isinstance(self.conv, MyConv2d) else filters
        y = F.conv2d(x, filters, None, self.stride, padding, self.dilation, in_channel)
        return y


class DynamicConv2d(nn.Module):
    def __init__(self, max_in_channels, max_out_channels, kernel_size=1, stride=1, dilation=1):
        super(DynamicConv2d, self).__init__()

        self.max_in_channels = max_in_channels
        self.max_out_channels = max_out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.dilation = dilation

        self.conv = nn.Conv2d(
            self.max_in_channels,
            self.max_out_channels,
            self.kernel_size,
            stride=self.stride,
            bias=False,
        )

        self.active_out_channel = self.max_out_channels

    def get_active_filter(self, out_channel, in_channel):
        return self.conv.weight[:out_channel, :in_channel, :, :]

    def forward(self, x, out_channel=None):
        if out_channel is None:
            out_channel = self.active_out_channel
        in_channel = x.size(1)
        filters = self.get_active_filter(out_channel, in_channel).contiguous()

        padding = get_same_padding(self.kernel_size)
        filters = self.conv.weight_standardization(filters) if isinstance(self.conv, MyConv2d) else filters
        y = F.conv2d(x, filters, None, self.stride, padding, self.dilation, 1)
        return y


class DynamicGroupConv2d(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size_list,
        groups_list,
        stride=1,
        dilation=1,
    ):
        super(DynamicGroupConv2d, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size_list = kernel_size_list
        self.groups_list = groups_list
        self.stride = stride
        self.dilation = dilation

        self.conv = nn.Conv2d(
            self.in_channels,
            self.out_channels,
            max(self.kernel_size_list),
            self.stride,
            groups=min(self.groups_list),
            bias=False,
        )

        self.active_kernel_size = max(self.kernel_size_list)
        self.active_groups = min(self.groups_list)

    def get_active_filter(self, kernel_size, groups):
        start, end = sub_filter_start_end(max(self.kernel_size_list), kernel_size)
        filters = self.conv.weight[:, :, start:end, start:end]

        sub_filters = torch.chunk(filters, groups, dim=0)
        sub_in_channels = self.in_channels // groups
        sub_ratio = filters.size(1) // sub_in_channels

        filter_crops = []
        for i, sub_filter in enumerate(sub_filters):
            part_id = i % sub_ratio
            start = part_id * sub_in_channels
            filter_crops.append(sub_filter[:, start : start + sub_in_channels, :, :])
        filters = torch.cat(filter_crops, dim=0)
        return filters

    def forward(self, x, kernel_size=None, groups=None):
        if kernel_size is None:
            kernel_size = self.active_kernel_size
        if groups is None:
            groups = self.active_groups

        filters = self.get_active_filter(kernel_size, groups).contiguous()
        padding = get_same_padding(kernel_size)
        filters = self.conv.weight_standardization(filters) if isinstance(self.conv, MyConv2d) else filters
        y = F.conv2d(
            x,
            filters,
            None,
            self.stride,
            padding,
            self.dilation,
            groups,
        )
        return y


class DynamicBatchNorm2d(nn.Module):
    SET_RUNNING_STATISTICS = False

    def __init__(self, max_feature_dim):
        super(DynamicBatchNorm2d, self).__init__()

        self.max_feature_dim = max_feature_dim
        self.bn = nn.BatchNorm2d(self.max_feature_dim)

    @staticmethod
    def bn_forward(x, bn: nn.BatchNorm2d, feature_dim):
        if bn.num_features == feature_dim or DynamicBatchNorm2d.SET_RUNNING_STATISTICS:
            return bn(x)
        else:
            exponential_average_factor = 0.0

            if bn.training and bn.track_running_stats:
                if bn.num_batches_tracked is not None:
                    bn.num_batches_tracked += 1
                    if bn.momentum is None:  # use cumulative moving average
                        exponential_average_factor = 1.0 / float(bn.num_batches_tracked)
                    else:  # use exponential moving average
                        exponential_average_factor = bn.momentum
            return F.batch_norm(
                x,
                bn.running_mean[:feature_dim],
                bn.running_var[:feature_dim],
                bn.weight[:feature_dim],
                bn.bias[:feature_dim],
                bn.training or not bn.track_running_stats,
                exponential_average_factor,
                bn.eps,
            )

    def forward(self, x):
        feature_dim = x.size(1)
        y = self.bn_forward(x, self.bn, feature_dim)
        return y


class DynamicGroupNorm(nn.GroupNorm):
    def __init__(self, num_groups, num_channels, eps=1e-5, affine=True, channel_per_group=None):
        super(DynamicGroupNorm, self).__init__(num_groups, num_channels, eps, affine)
        self.channel_per_group = channel_per_group

    def forward(self, x):
        n_channels = x.size(1)
        n_groups = n_channels // self.channel_per_group
        return F.group_norm(x, n_groups, self.weight[:n_channels], self.bias[:n_channels], self.eps)

    @property
    def bn(self):
        return self


class DynamicSE(SEModule):
    def __init__(self, max_channel):
        super(DynamicSE, self).__init__(max_channel)

    def get_active_reduce_weight(self, num_mid, in_channel, groups=None):
        if groups is None or groups == 1:
            return self.fc.reduce.weight[:num_mid, :in_channel, :, :]
        else:
            assert in_channel % groups == 0
            sub_in_channels = in_channel // groups
            sub_filters = torch.chunk(self.fc.reduce.weight[:num_mid, :, :, :], groups, dim=1)
            return torch.cat(
                [sub_filter[:, :sub_in_channels, :, :] for sub_filter in sub_filters],
                dim=1,
            )

    def get_active_reduce_bias(self, num_mid):
        return self.fc.reduce.bias[:num_mid] if self.fc.reduce.bias is not None else None

    def get_active_expand_weight(self, num_mid, in_channel, groups=None):
        if groups is None or groups == 1:
            return self.fc.expand.weight[:in_channel, :num_mid, :, :]
        else:
            assert in_channel % groups == 0
            sub_in_channels = in_channel // groups
            sub_filters = torch.chunk(self.fc.expand.weight[:, :num_mid, :, :], groups, dim=0)
            return torch.cat(
                [sub_filter[:sub_in_channels, :, :, :] for sub_filter in sub_filters],
                dim=0,
            )

    def get_active_expand_bias(self, in_channel, groups=None):
        if groups is None or groups == 1:
            return self.fc.expand.bias[:in_channel] if self.fc.expand.bias is not None else None
        else:
            assert in_channel % groups == 0
            sub_in_channels = in_channel // groups
            sub_bias_list = torch.chunk(self.fc.expand.bias, groups, dim=0)
            return torch.cat([sub_bias[:sub_in_channels] for sub_bias in sub_bias_list], dim=0)

    def forward(self, x, groups=None):
        in_channel = x.size(1)
        num_mid = make_divisible(in_channel // self.reduction, divisor=MyNetwork.CHANNEL_DIVISIBLE)

        y = x.mean(3, keepdim=True).mean(2, keepdim=True)
        # reduce
        reduce_filter = self.get_active_reduce_weight(num_mid, in_channel, groups=groups).contiguous()
        reduce_bias = self.get_active_reduce_bias(num_mid)
        y = F.conv2d(y, reduce_filter, reduce_bias, 1, 0, 1, 1)
        # relu
        y = self.fc.relu(y)
        # expand
        expand_filter = self.get_active_expand_weight(num_mid, in_channel, groups=groups).contiguous()
        expand_bias = self.get_active_expand_bias(in_channel, groups=groups)
        y = F.conv2d(y, expand_filter, expand_bias, 1, 0, 1, 1)
        # hard sigmoid
        y = self.fc.h_sigmoid(y)

        return x * y


class DynamicLinear(nn.Module):
    def __init__(self, max_in_features, max_out_features, bias=True):
        super(DynamicLinear, self).__init__()

        self.max_in_features = max_in_features
        self.max_out_features = max_out_features
        self.bias = bias

        self.linear = nn.Linear(self.max_in_features, self.max_out_features, self.bias)

        self.active_out_features = self.max_out_features

    def get_active_weight(self, out_features, in_features):
        return self.linear.weight[:out_features, :in_features]

    def get_active_bias(self, out_features):
        return self.linear.bias[:out_features] if self.bias else None

    def forward(self, x, out_features=None):
        if out_features is None:
            out_features = self.active_out_features

        in_features = x.size(1)
        weight = self.get_active_weight(out_features, in_features).contiguous()
        bias = self.get_active_bias(out_features)
        y = F.linear(x, weight, bias)
        return y



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys

from dynast.search.search_tactic import LINAS, Evolutionary, LINASDistributed, RandomSearch, RandomSearchDistributed
from dynast.utils import check_kwargs_deprecated, log, set_logger
from dynast.utils.distributed import get_distributed_vars, init_distributed


class DyNAS:
    '''
    The DyNAS class manages the user input configuration parameters, the search tactics, and
    the supernetwork configuration. Since the search tactics vary widely, they are instantiate
    with this class based on the user input.
    '''

    def __new__(self, **kwargs):
        log_level = logging.INFO
        if kwargs.get('verbose'):
            log_level = logging.DEBUG
        set_logger(level=log_level)

        LOCAL_RANK, WORLD_RANK, WORLD_SIZE, DIST_METHOD = get_distributed_vars()
        if DIST_METHOD:
            backend = kwargs.get('backend', 'gloo')
            init_distributed(backend, WORLD_RANK, WORLD_SIZE)
            seed = kwargs.get('seed', None)
            if seed:
                kwargs['seed'] = seed + WORLD_RANK

        log.info('=' * 40)
        log.info('Starting Dynamic NAS Toolkit (DyNAS-T)')
        log.info('=' * 40)

        # Required arguments for the DyNAS class
        # TODO(macsz) If these are common for all created classes, then we can move it out of kwargs
        REQUIRED_KWARGS = [
            'supernet',
            'optimization_metrics',
            'measurements',
            'search_tactic',
            'num_evals',
            'results_path',
            'dataset_path',
        ]

        # Validity checks
        for argument in REQUIRED_KWARGS:
            if argument not in kwargs:
                log.error(f"Missing `--{argument}` parameter.")
                sys.exit("Missing argument, see log file for info.")

        kwargs = check_kwargs_deprecated(**kwargs)

        if len(kwargs['optimization_metrics']) > 3:
            log.error('Number of optimization_metrics is out of range. 1-3 supported.')
            sys.exit()

        log.info('-' * 40)
        log.info('DyNAS Parameter Inputs:')
        for key, value in kwargs.items():
            log.info(f'{key}: {value}')
        log.info('-' * 40)

        if kwargs.get('distributed', False):
            # LINAS bi-level evolutionary algorithm search distributed to multiple workers
            if kwargs['search_tactic'] == 'linas':
                log.info('Initializing DyNAS LINAS (distributed) algorithm object.')
                return LINASDistributed(**kwargs)

                # Uniform random sampling of the architectural space distributed to multiple workers
            elif kwargs['search_tactic'] == 'random':
                log.info('Initializing DyNAS random (distributed) search algorithm object.')
                return RandomSearchDistributed(**kwargs)

        # LINAS bi-level evolutionary algorithm search
        if kwargs['search_tactic'] == 'linas':
            log.info('Initializing DyNAS LINAS algorithm object.')
            return LINAS(**kwargs)

        # Standard evolutionary algorithm search
        elif kwargs['search_tactic'] == 'evolutionary':
            log.info('Initializing DyNAS evoluationary algorithm object.')
            return Evolutionary(**kwargs)

        # Uniform random sampling of the architectural space
        elif kwargs['search_tactic'] == 'random':
            log.info('Initializing DyNAS random search algorithm object.')
            return RandomSearch(**kwargs)

        else:
            error_message = (
                "Invalid `--search_tactic` parameter `{}` (options: 'linas', 'evolutionary', 'random').".format(
                    kwargs['search_tactic']
                )
            )  # TODO(macsz) Un-hardcode options.
            log.error(error_message)
            raise NotImplementedError(error_message)



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle

import numpy as np
from scipy.stats import kendalltau, spearmanr
from sklearn import linear_model, svm
from sklearn.metrics import mean_absolute_percentage_error, mean_squared_error
from sklearn.model_selection import GridSearchCV


class Predictor:
    DEFAULT_ALPHAS = np.arange(0.1, 10.1, 0.1)
    DEFAULT_COST_FACTORS = np.arange(1.0, 101.0, 1.0)
    DEFAULT_MAX_ITERATIONS = 1000000

    def __init__(
        self,
        alphas=DEFAULT_ALPHAS,
        cost_factors=DEFAULT_COST_FACTORS,
        max_iterations=DEFAULT_MAX_ITERATIONS,
        verbose=False,
    ):
        SEARCHER_VERBOSITY = 10

        # Initialize label normalization factor
        self.normalization_factor = 1.0

        # Initialize best searcher index
        self.best_index = 0

        # Create lists of regressors and associated hyper-parameters
        regressors = [
            linear_model.Ridge(max_iter=max_iterations),
            svm.SVR(kernel='rbf', gamma='auto', epsilon=0.0, max_iter=max_iterations),
        ]
        hyper_parameters = [{'alpha': alphas}, {'C': cost_factors}]

        # Create list of hyper-parameter searchers
        self.searchers = []
        for regressor, parameters in zip(regressors, hyper_parameters):
            self.searchers.append(
                GridSearchCV(
                    estimator=regressor,
                    param_grid=parameters,
                    n_jobs=-1,
                    scoring='neg_mean_absolute_percentage_error',
                    verbose=SEARCHER_VERBOSITY if (verbose) else 0,
                )
            )

    def train(self, examples, labels):
        '''
        Trains the predictor on the specified examples and labels using the underlying regressor.

        Parameters
        ----------
            examples: Examples to be used for training.
            labels: Labels to be used for training.

        Returns
        -------
            None
        '''

        # Compute normalization factor
        max_label = np.amax(np.abs(labels))
        if max_label > 0.0:
            self.normalization_factor = 10 ** (np.floor(np.log10(max_label)) - 1.0)
        else:
            self.normalization_factor = 1.0

        # Compute normalized labels
        normalized_labels = labels / self.normalization_factor

        # Train regressors with optimal parameters
        scores = np.zeros(len(self.searchers))
        for s in range(len(self.searchers)):
            self.searchers[s].fit(examples, normalized_labels)
            scores[s] = self.searchers[s].best_score_

        # Determine index of best searcher
        self.best_index = np.argmax(scores)

    def predict(self, examples):
        '''
        Predicts the output values of the specified examples using the underlying regressor.

        Parameters
        ----------
            examples: Examples for which predictions will be made.

        Returns
        -------
            Predictions of the specified examples.
        '''

        # Compute predictions
        regressor = self.searchers[self.best_index].best_estimator_
        normalized_predictions = regressor.predict(examples)
        predictions = normalized_predictions * self.normalization_factor

        return predictions

    def get_parameters(self):
        '''
        Returns the optimal parameter values of the underlying regressor.

        Parameters
        ----------
            None

        Returns
        -------
            Optimal parameter values of the underlying regressor.
        '''

        # Retrieve optimal parameters
        parameters = {'best_index': self.best_index}
        for searcher in self.searchers:
            regressor_name = searcher.best_estimator_.__class__.__name__
            for key in searcher.best_params_:
                parameter_key = regressor_name + '_' + key
                parameters[parameter_key.lower()] = searcher.best_params_[key]

        return parameters

    def get_metrics(self, examples, labels):
        '''
        Computes the performance metrics of the underlying regressor.

        Parameters
        ----------
            examples: Examples to use when computing performance metrics.

            labels: Labels to use when computing performance metrics.

        Returns
        -------
            Performance metrics of the underlying regressor. The metrics are
                Mean absolute percentage error (MAPE)
                Root mean squared error (RMSE)
                Kendall rank correlation coefficient (kendall)
                Spearman's rank correlation coefficient (spearman)
        '''

        # Compute predictions of specified examples
        predictions = self.predict(examples)

        # Compute performance metrics
        mape = 100.0 * mean_absolute_percentage_error(labels, predictions)
        rmse = mean_squared_error(labels, predictions, squared=False)
        kendall, _ = kendalltau(labels, predictions)
        spearman, _ = spearmanr(labels, predictions)

        return mape, rmse, kendall, spearman

    def load(self, filename):
        '''
        Loads the model of the underlying regressor and searcher.

        Parameters
        ----------
            filename: Name of the file from which to load the model.

        Returns
        -------
            None
        '''

        # Load searcher and regressor from specified file
        with open(filename, 'rb') as input_file:
            self.normalization_factor = pickle.load(input_file)
            self.best_index = pickle.load(input_file)
            self.searchers = pickle.load(input_file)

    def save(self, filename):
        '''
        Saves the model of the underlying regressor and searcher.

        Parameters
        ----------
            filename: Name of the file to which to save the model.

        Returns
        -------
            None
        '''

        # Save searcher and regressor to specified file
        with open(filename, 'wb') as output_file:
            pickle.dump(self.normalization_factor, output_file)
            pickle.dump(self.best_index, output_file)
            pickle.dump(self.searchers, output_file)



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This software is subject to the terms and conditions entered into between the parties.

import ast
import random
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

from dynast.utils import log


class EncodingBase:
    """This class manages various representations of the sub-network configuration. It processes
    a user parameter dictionary to create the following representations:
    """

    def __init__(self, param_dict: dict, verbose: bool = False, seed: int = 0):
        self.param_dict = param_dict
        self.verbose = verbose
        self.mapper, self.param_upperbound, self.param_count = self.process_param_dict()
        self.set_seed(seed)
        self.inv_mapper = self.create_inv_mapper()

    def process_param_dict(self) -> Tuple[List[Dict[int, Union[int, float]]], List[int], int]:
        """Builds a parameter mapping arrays and an upper-bound vector for PyMoo."""
        parameter_count = 0
        parameter_bound = list()
        parameter_upperbound = list()
        parameter_mapper = list()

        for parameter, options in self.param_dict.items():
            # How many variables should be searched for
            parameter_count += options['count']
            parameter_bound.append(options['count'])

            # How many variables for each parameter
            for i in range(options['count']):
                parameter_upperbound.append(len(options['vars']) - 1)
                single_mapping = dict()
                for idx, value in enumerate(options['vars']):
                    if type(value) == int or type(value) == float:
                        single_mapping[idx] = value
                    else:
                        single_mapping[idx] = str(value)

                parameter_mapper.append(single_mapping)

        if self.verbose:
            log.info('Problem definition variables: {}'.format(parameter_count))
            log.info('Variable Upper Bound array: {}'.format(parameter_upperbound))
            log.info('Mapping dictionary created of length: {}'.format(len(parameter_mapper)))
            log.info('Parameter Bound: {}'.format(parameter_bound))

        return parameter_mapper, parameter_upperbound, parameter_count

    def create_inv_mapper(self) -> List[Dict[int, int]]:
        """Builds inverse of self.mapper."""
        inv_parameter_mapper = list()

        for value in self.mapper:
            inverse_dict = {v: k for k, v in value.items()}
            inv_parameter_mapper.append(inverse_dict)

        return inv_parameter_mapper

    def onehot_generic(self, in_array: List[int]) -> np.ndarray:
        """This is a generic approach to one-hot vectorization for predictor training and testing.

        It does not account for unused parameter mapping (e.g. block depth).
        For unused parameter mapping, the end user will need to provide a custom solution.

        input_array - the pymoo individual 1-D vector
        mapper - the map for elastic parameters of the supernetwork
        """
        # Insure compatible array and mapper
        assert len(in_array) == len(self.mapper)

        onehot = list()

        # This function converts a pymoo input vector to a one-hot feature vector
        for i in range(len(self.mapper)):
            segment = [0 for _ in range(len(self.mapper[i]))]
            segment[in_array[i]] = 1
            onehot.extend(segment)

        return np.array(onehot)

    def random_sample(self) -> List[int]:
        """Generates a random subnetwork from the possible elastic parameter range."""

        pymoo_vector = list()
        for i in range(len(self.mapper)):
            options = [x for x in range(len(self.mapper[i]))]
            pymoo_vector.append(random.choice(options))

        return pymoo_vector

    def random_samples(self, size: int = 100, trial_limit: int = 100000) -> List[List[int]]:
        """Generates a list of random subnetworks from the possible elastic parameter range."""
        pymoo_vector_list: List[List[int]] = list()

        trials = 0
        while len(pymoo_vector_list) < size and trials < trial_limit:
            sample = self.random_sample()
            if sample not in pymoo_vector_list:
                pymoo_vector_list.append(sample)
            trials += 1

        if trials >= trial_limit:
            log.warning('Unable to create unique list of samples.')

        return pymoo_vector_list

    def translate2param(self, pymoo_vector: List[int]) -> Dict['str', List[int]]:
        """Translate a PyMoo 1-D parameter vector back to the elastic parameter dictionary format."""
        output: Dict['str', List[int]] = dict()

        # Assign (and map) each vector element to the appropriate parameter dictionary key
        counter = 0
        for key, value in self.param_dict.items():
            output[key] = list()
            for i in range(value['count']):
                # Note: following round term was added to support CMA-ES
                output[key].append(self.mapper[counter][round(pymoo_vector[counter])])
                counter += 1

        # Insure correct vector mapping occurred
        assert counter == len(self.mapper)

        return output

    def translate2pymoo(self, parameters: Dict['str', List[int]]):
        """Translate a single parameter dict to pymoo vector."""

        output = list()

        mapper_counter = 0
        for key, value in self.param_dict.items():
            param_counter = 0
            for i in range(value['count']):
                output.append(self.inv_mapper[mapper_counter][parameters[key][param_counter]])
                mapper_counter += 1
                param_counter += 1

        return output

    def import_csv(
        self,
        filepath: str,
        config: str,
        objective: str,
        column_names: Optional[List[str]] = None,
        drop_duplicates: bool = True,
    ) -> pd.DataFrame:
        """Import a csv file generated from a supernetwork search for the purpose of training a predictor.

        Parameters
        ----------
        * filepath - path of the csv to be imported.
        * config - column name with the subnetwork configuration.
        * objective - target/label for the subnet configuration (e.g. accuracy, latency).
        * column_names - a list of column names for the dataframe.
        * drop_duplicates - whether to drop duplicate entries.

        Returns
        -------
        * df - the output dataframe that contains the original config dict, pymoo, and 1-hot equivalent vector for training.
        """

        if column_names == None:
            df = pd.read_csv(filepath)
        else:
            df = pd.read_csv(filepath)
            df.columns = column_names
        df = df[[config, objective]]

        # Old corner case coverage
        df[config] = df[config].replace({'null': 'None'}, regex=True)

        if drop_duplicates:
            df.drop_duplicates(subset=[config], inplace=True)
            df.reset_index(drop=True, inplace=True)

        convert_to_dict = list()
        convert_to_pymoo = list()
        convert_to_onehot = list()
        for i in range(len(df)):
            # Elastic Param Config format
            config_as_dict = ast.literal_eval(df[config].iloc[i])
            convert_to_dict.append(config_as_dict)
            # PyMoo 1-D vector format
            config_as_pymoo = self.translate2pymoo(config_as_dict)
            convert_to_pymoo.append(config_as_pymoo)
            # Onehot predictor format
            config_as_onehot = self.onehot_generic(config_as_pymoo)
            convert_to_onehot.append(config_as_onehot)

        df[config] = convert_to_dict
        df['config_pymoo'] = convert_to_pymoo
        df['config_onehot'] = convert_to_onehot

        return df

    def set_seed(self, seed: int) -> None:
        """Set the random seed for randomized subnet generation and test/train split"""
        self.seed = seed
        random.seed(seed)

    @staticmethod
    def create_training_set(
        dataframe: pd.DataFrame,
        config: str = 'subnet',
        train_with_all: bool = True,
        split: float = 0.33,
        seed: Optional[int] = None,
    ):
        """Create a sklearn compatible test/train set from an imported results csv
        after "import_csv" method is run.
        """

        collect_rows = list()
        for i in range(len(dataframe)):
            collect_rows.append(np.asarray(dataframe['config_onehot'].iloc[i]))
        features = np.asarray(collect_rows)

        labels = dataframe.drop(columns=[config, 'config_pymoo', 'config_onehot']).values

        assert len(features) == len(labels)

        if train_with_all:
            log.info('Training set length={}'.format(len(labels)))
            return features, labels
        else:
            features_train, features_test, labels_train, labels_test = train_test_split(
                features, labels, test_size=split, random_state=seed
            )
            log.info('Test ({}) Train ({}) ratio is {}.'.format(len(labels_train), len(labels_test), split))
            return features_train, features_test, labels_train, labels_test



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Once for All: Train One Network and Specialize it for Efficient Deployment
# Han Cai, Chuang Gan, Tianzhe Wang, Zhekai Zhang, Song Han
# International Conference on Learning Representations (ICLR), 2020.

import json
import os
import random
import time

import numpy as np
import torch.backends.cudnn as cudnn
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.parallel
import torch.optim

from dynast.supernetwork.image_classification.ofa.ofa.utils import (
    MyRandomResizedCrop,
    cross_entropy_loss_with_soft_target,
    cross_entropy_with_label_smoothing,
    get_net_info,
    init_models,
    mix_images,
    mix_labels,
    write_log,
)
from dynast.utils.nn import AverageMeter, accuracy

__all__ = ["RunManager"]


class RunManager:
    def __init__(
        self,
        path,
        net,
        run_config,
        init=True,
        measure_latency=None,
        no_gpu=False,
        verbose=False,
    ):
        self.path = path
        self.net = net
        self.run_config = run_config
        self.verbose = verbose

        self.best_acc = 0
        self.start_epoch = 0

        os.makedirs(self.path, exist_ok=True)

        # move network to GPU if available
        if torch.cuda.is_available() and (not no_gpu):
            self.device = torch.device("cuda:0")
            self.net = self.net.to(self.device)
            cudnn.benchmark = True
        else:
            self.device = torch.device("cpu")
        # initialize model (default)
        if init:
            init_models(run_config.model_init)

        # net info
        net_info = get_net_info(self.net, self.run_config.data_provider.data_shape, measure_latency)
        with open("%s/net_info.txt" % self.path, "w") as fout:
            fout.write(json.dumps(net_info, indent=4) + "\n")
            # noinspection PyBroadException
            try:
                fout.write(self.network.module_str + "\n")
            except Exception:
                pass
            fout.write("%s\n" % self.run_config.data_provider.train.dataset.transform)
            fout.write("%s\n" % self.run_config.data_provider.test.dataset.transform)
            fout.write("%s\n" % self.network)

        # criterion
        if isinstance(self.run_config.mixup_alpha, float):
            self.train_criterion = cross_entropy_loss_with_soft_target
        elif self.run_config.label_smoothing > 0:
            self.train_criterion = lambda pred, target: cross_entropy_with_label_smoothing(
                pred, target, self.run_config.label_smoothing
            )
        else:
            self.train_criterion = nn.CrossEntropyLoss()
        self.test_criterion = nn.CrossEntropyLoss()

        # optimizer
        if self.run_config.no_decay_keys:
            keys = self.run_config.no_decay_keys.split("#")
            net_params = [
                self.network.get_parameters(keys, mode="exclude"),  # parameters with weight decay
                self.network.get_parameters(keys, mode="include"),  # parameters without weight decay
            ]
        else:
            # noinspection PyBroadException
            try:
                net_params = self.network.weight_parameters()
            except Exception:
                net_params = []
                for param in self.network.parameters():
                    if param.requires_grad:
                        net_params.append(param)
        self.optimizer = self.run_config.build_optimizer(net_params)

        self.net = torch.nn.DataParallel(self.net)

    """ save path and log path """

    @property
    def save_path(self):
        if self.__dict__.get("_save_path", None) is None:
            save_path = os.path.join(self.path, "checkpoint")
            os.makedirs(save_path, exist_ok=True)
            self.__dict__["_save_path"] = save_path
        return self.__dict__["_save_path"]

    @property
    def logs_path(self):
        if self.__dict__.get("_logs_path", None) is None:
            logs_path = os.path.join(self.path, "logs")
            os.makedirs(logs_path, exist_ok=True)
            self.__dict__["_logs_path"] = logs_path
        return self.__dict__["_logs_path"]

    @property
    def network(self):
        return self.net.module if isinstance(self.net, nn.DataParallel) else self.net

    def write_log(self, log_str, prefix="valid", should_print=True, mode="a"):
        write_log(self.logs_path, log_str, prefix, should_print, mode)

    """ save and load models """

    def save_model(self, checkpoint=None, is_best=False, model_name=None):
        if checkpoint is None:
            checkpoint = {"state_dict": self.network.state_dict()}

        if model_name is None:
            model_name = "checkpoint.pth.tar"

        checkpoint["dataset"] = self.run_config.dataset  # add `dataset` info to the checkpoint
        latest_fname = os.path.join(self.save_path, "latest.txt")
        model_path = os.path.join(self.save_path, model_name)
        with open(latest_fname, "w") as fout:
            fout.write(model_path + "\n")
        torch.save(checkpoint, model_path)

        if is_best:
            best_path = os.path.join(self.save_path, "model_best.pth.tar")
            torch.save({"state_dict": checkpoint["state_dict"]}, best_path)

    def load_model(self, model_fname=None):
        latest_fname = os.path.join(self.save_path, "latest.txt")
        if model_fname is None and os.path.exists(latest_fname):
            with open(latest_fname, "r") as fin:
                model_fname = fin.readline()
                if model_fname[-1] == "\n":
                    model_fname = model_fname[:-1]
        # noinspection PyBroadException
        try:
            if model_fname is None or not os.path.exists(model_fname):
                model_fname = "%s/checkpoint.pth.tar" % self.save_path
                with open(latest_fname, "w") as fout:
                    fout.write(model_fname + "\n")
            print("=> loading checkpoint '{}'".format(model_fname))
            checkpoint = torch.load(model_fname, map_location="cpu")
        except Exception:
            print("fail to load checkpoint from %s" % self.save_path)
            return {}

        self.network.load_state_dict(checkpoint["state_dict"])
        if "epoch" in checkpoint:
            self.start_epoch = checkpoint["epoch"] + 1
        if "best_acc" in checkpoint:
            self.best_acc = checkpoint["best_acc"]
        if "optimizer" in checkpoint:
            self.optimizer.load_state_dict(checkpoint["optimizer"])

        print("=> loaded checkpoint '{}'".format(model_fname))
        return checkpoint

    def save_config(self, extra_run_config=None, extra_net_config=None):
        """dump run_config and net_config to the model_folder"""
        run_save_path = os.path.join(self.path, "run.config")
        if not os.path.isfile(run_save_path):
            run_config = self.run_config.config
            if extra_run_config is not None:
                run_config.update(extra_run_config)
            json.dump(run_config, open(run_save_path, "w"), indent=4)
            print("Run configs dump to %s" % run_save_path)

        try:
            net_save_path = os.path.join(self.path, "net.config")
            net_config = self.network.config
            if extra_net_config is not None:
                net_config.update(extra_net_config)
            json.dump(net_config, open(net_save_path, "w"), indent=4)
            print("Network configs dump to %s" % net_save_path)
        except Exception:
            print("%s do not support net config" % type(self.network))

    """ metric related """

    def get_metric_dict(self):
        return {
            "top1": AverageMeter(),
            "top5": AverageMeter(),
        }

    def update_metric(self, metric_dict, output, labels):
        acc1, acc5 = accuracy(output, labels, topk=(1, 5))
        metric_dict["top1"].update(acc1, output.size(0))
        metric_dict["top5"].update(acc5, output.size(0))

    def get_metric_vals(self, metric_dict, return_dict=False):
        if return_dict:
            return {key: metric_dict[key].avg for key in metric_dict}
        else:
            return [metric_dict[key].avg for key in metric_dict]

    def get_metric_names(self):
        return "top1", "top5"

    """ train and test """

    def validate(
        self,
        epoch=0,
        is_test=False,
        run_str="",
        net=None,
        data_loader=None,
        no_logs=False,
        train_mode=False,
    ):
        if net is None:
            net = self.net
        if not isinstance(net, nn.DataParallel):
            net = nn.DataParallel(net)

        if data_loader is None:
            data_loader = self.run_config.test_loader if is_test else self.run_config.valid_loader

        if train_mode:
            net.train()
        else:
            net.eval()

        losses = AverageMeter()
        metric_dict = self.get_metric_dict()

        with torch.no_grad():
            for i, (images, labels) in enumerate(data_loader):
                images, labels = images.to(self.device), labels.to(self.device)
                # compute output
                output = net(images)
                loss = self.test_criterion(output, labels)
                # measure accuracy and record loss
                self.update_metric(metric_dict, output, labels)

                losses.update(loss.item(), images.size(0))
        return losses.avg, self.get_metric_vals(metric_dict)

    def validate_all_resolution(self, epoch=0, is_test=False, net=None):
        if net is None:
            net = self.network
        if isinstance(self.run_config.data_provider.image_size, list):
            img_size_list, loss_list, top1_list, top5_list = [], [], [], []
            for img_size in self.run_config.data_provider.image_size:
                img_size_list.append(img_size)
                self.run_config.data_provider.assign_active_img_size(img_size)
                self.reset_running_statistics(net=net)
                loss, (top1, top5) = self.validate(epoch, is_test, net=net)
                loss_list.append(loss)
                top1_list.append(top1)
                top5_list.append(top5)
            return img_size_list, loss_list, top1_list, top5_list
        else:
            loss, (top1, top5) = self.validate(epoch, is_test, net=net)
            return (
                [self.run_config.data_provider.active_img_size],
                [loss],
                [top1],
                [top5],
            )

    def train_one_epoch(self, args, epoch, warmup_epochs=0, warmup_lr=0):
        # switch to train mode
        self.net.train()
        MyRandomResizedCrop.EPOCH = epoch  # required by elastic resolution

        nBatch = len(self.run_config.train_loader)

        losses = AverageMeter()
        metric_dict = self.get_metric_dict()
        data_time = AverageMeter()

        end = time.time()
        for i, (images, labels) in enumerate(self.run_config.train_loader):
            MyRandomResizedCrop.BATCH = i
            data_time.update(time.time() - end)
            if epoch < warmup_epochs:
                new_lr = self.run_config.warmup_adjust_learning_rate(
                    self.optimizer,
                    warmup_epochs * nBatch,
                    nBatch,
                    epoch,
                    i,
                    warmup_lr,
                )
            else:
                new_lr = self.run_config.adjust_learning_rate(self.optimizer, epoch - warmup_epochs, i, nBatch)

            images, labels = images.to(self.device), labels.to(self.device)
            target = labels
            if isinstance(self.run_config.mixup_alpha, float):
                # transform data
                lam = random.betavariate(self.run_config.mixup_alpha, self.run_config.mixup_alpha)
                images = mix_images(images, lam)
                labels = mix_labels(
                    labels,
                    lam,
                    self.run_config.data_provider.n_classes,
                    self.run_config.label_smoothing,
                )

            # soft target
            if args.teacher_model is not None:
                args.teacher_model.train()
                with torch.no_grad():
                    soft_logits = args.teacher_model(images).detach()
                    soft_label = F.softmax(soft_logits, dim=1)

            # compute output
            output = self.net(images)
            loss = self.train_criterion(output, labels)

            if args.teacher_model is None:
                loss_type = "ce"
            else:
                if args.kd_type == "ce":
                    kd_loss = cross_entropy_loss_with_soft_target(output, soft_label)
                else:
                    kd_loss = F.mse_loss(output, soft_logits)
                loss = args.kd_ratio * kd_loss + loss
                loss_type = "%.1fkd+ce" % args.kd_ratio

            # compute gradient and do SGD step
            self.net.zero_grad()  # or self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

            # measure accuracy and record loss
            losses.update(loss.item(), images.size(0))
            self.update_metric(metric_dict, output, target)

            end = time.time()
        return losses.avg, self.get_metric_vals(metric_dict)

    def train(self, args, warmup_epoch=0, warmup_lr=0):
        for epoch in range(self.start_epoch, self.run_config.n_epochs + warmup_epoch):
            train_loss, (train_top1, train_top5) = self.train_one_epoch(args, epoch, warmup_epoch, warmup_lr)

            if (epoch + 1) % self.run_config.validation_frequency == 0:
                img_size, val_loss, val_acc, val_acc5 = self.validate_all_resolution(epoch=epoch, is_test=False)

                is_best = np.mean(val_acc) > self.best_acc
                self.best_acc = max(self.best_acc, np.mean(val_acc))
                val_log = "Valid [{0}/{1}]\tloss {2:.3f}\t{5} {3:.3f} ({4:.3f})".format(
                    epoch + 1 - warmup_epoch,
                    self.run_config.n_epochs,
                    np.mean(val_loss),
                    np.mean(val_acc),
                    self.best_acc,
                    self.get_metric_names()[0],
                )
                val_log += "\t{2} {0:.3f}\tTrain {1} {top1:.3f}\tloss {train_loss:.3f}\t".format(
                    np.mean(val_acc5), *self.get_metric_names(), top1=train_top1, train_loss=train_loss
                )
                for i_s, v_a in zip(img_size, val_acc):
                    val_log += "(%d, %.3f), " % (i_s, v_a)
                self.write_log(val_log, prefix="valid", should_print=False)
            else:
                is_best = False

            self.save_model(
                {
                    "epoch": epoch,
                    "best_acc": self.best_acc,
                    "optimizer": self.optimizer.state_dict(),
                    "state_dict": self.network.state_dict(),
                },
                is_best=is_best,
            )

    def reset_running_statistics(self, net=None, subset_size=2000, subset_batch_size=200, data_loader=None):
        from dynast.supernetwork.image_classification.ofa.ofa.imagenet_classification.elastic_nn.utils import (
            set_running_statistics,
        )

        if net is None:
            net = self.network
        if data_loader is None:
            data_loader = self.run_config.random_sub_train_loader(subset_size, subset_batch_size)
        set_running_statistics(net, data_loader)



# Copyright 2018- The Hugging Face team. All rights reserved.
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from torch import nn
from transformers.modeling_outputs import BaseModelOutputWithPastAndCrossAttentions
from transformers.models.bert.modeling_bert import (
    BertAttention,
    BertEncoder,
    BertForSequenceClassification,
    BertIntermediate,
    BertLayer,
    BertModel,
    BertOutput,
    BertPooler,
    BertSelfAttention,
    BertSelfOutput,
)

from ..machine_translation.modules_supernetwork import LinearSuper


class BertSupernetSelfAttention(BertSelfAttention):
    def __init__(self, config):
        super().__init__(config)
        self.query = LinearSuper(config.hidden_size, self.all_head_size)
        self.key = LinearSuper(config.hidden_size, self.all_head_size)
        self.value = LinearSuper(config.hidden_size, self.all_head_size)

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (self.subnet_attention_heads, self.attention_head_size)
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def set_sample_config(self, sub_layer_config):
        self.subnet_attention_heads = sub_layer_config['num_attention_heads']
        self.subnet_hidden_size = sub_layer_config['hidden_size']
        self.all_head_size = self.attention_head_size * self.subnet_attention_heads

        self.query.set_sample_config(self.subnet_hidden_size, self.all_head_size)
        self.key.set_sample_config(self.subnet_hidden_size, self.all_head_size)
        self.value.set_sample_config(self.subnet_hidden_size, self.all_head_size)


class BertSupernetSelfOutput(BertSelfOutput):
    def __init__(self, config):
        super().__init__(config)

        self.supernet_attention_head_size = int(config.hidden_size / config.num_attention_heads)
        self.dense = LinearSuper(config.hidden_size, config.hidden_size)

    def set_sample_config(self, sub_layer_config):
        subnet_all_head_size = self.supernet_attention_head_size * sub_layer_config['num_attention_heads']
        self.dense.set_sample_config(subnet_all_head_size, sub_layer_config['hidden_size'])


class BertSupernetAttention(BertAttention):
    def __init__(self, config):
        super().__init__(config)
        self.self = BertSupernetSelfAttention(config)
        self.output = BertSupernetSelfOutput(config)

    def set_sample_config(self, sub_layer_config):
        self.self.set_sample_config(sub_layer_config)
        self.output.set_sample_config(sub_layer_config)


class BertSupernetIntermediate(BertIntermediate):
    def __init__(self, config):
        super().__init__(config)
        self.dense = LinearSuper(config.hidden_size, config.intermediate_size)

    def set_sample_config(self, sub_layer_config):
        self.dense.set_sample_config(sub_layer_config['hidden_size'], sub_layer_config['intermediate_size'])


class BertSupernetOutput(BertOutput):
    def __init__(self, config):
        super().__init__(config)
        self.dense = LinearSuper(config.intermediate_size, config.hidden_size)

    def set_sample_config(self, sub_layer_config):
        self.dense.set_sample_config(sub_layer_config['intermediate_size'], sub_layer_config['hidden_size'])


class BertSupernetLayer(BertLayer):
    def __init__(self, config):
        super().__init__(config)
        self.attention = BertSupernetAttention(config)
        self.intermediate = BertSupernetIntermediate(config)
        self.output = BertSupernetOutput(config)

    def set_sample_config(self, sub_layer_config):
        self.attention.set_sample_config(sub_layer_config)
        self.intermediate.set_sample_config(sub_layer_config)
        self.output.set_sample_config(sub_layer_config)


class BertSupernetEncoder(BertEncoder):
    def __init__(self, config):
        super().__init__(config)
        self.layer = nn.ModuleList([BertSupernetLayer(config) for _ in range(config.num_hidden_layers)])

    def forward(
        self,
        hidden_states,
        attention_mask=None,
        head_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        past_key_values=None,
        use_cache=None,
        output_attentions=False,
        output_hidden_states=False,
        return_dict=True,
    ):
        for layer_idx, layer_module in enumerate(self.layer):
            if layer_idx >= self.subnet_num_layers:
                break

            layer_head_mask = None
            past_key_value = None

            layer_outputs = layer_module(
                hidden_states,
                attention_mask,
                layer_head_mask,
                encoder_hidden_states,
                encoder_attention_mask,
                past_key_value,
                output_attentions,
            )

            hidden_states = layer_outputs[0]

        return BaseModelOutputWithPastAndCrossAttentions(
            last_hidden_state=hidden_states,
        )

    def set_sample_config(self, subnetwork_config):
        self.subnet_num_layers = subnetwork_config["num_layers"]
        for layer_idx in range(self.subnet_num_layers):
            sub_layer_config = {
                'intermediate_size': subnetwork_config['intermediate_size'][layer_idx],
                'num_attention_heads': subnetwork_config['num_attention_heads'][layer_idx],
                'hidden_size': subnetwork_config["hidden_size"],
            }
            self.layer[layer_idx].set_sample_config(sub_layer_config)


class BertSupernetPooler(BertPooler):
    def __init__(self, config):
        super().__init__(config)
        self.dense = LinearSuper(config.hidden_size, config.hidden_size)

    def set_sample_config(self, subnetwork_config):
        self.dense.set_sample_config(subnetwork_config["hidden_size"], subnetwork_config["hidden_size"])


class BertSupernetModel(BertModel):
    def __init__(self, config):
        super().__init__(config)
        self.config = config
        self.encoder = BertSupernetEncoder(config)
        self.pooler = BertSupernetPooler(config)

    def set_sample_config(self, subnetwork_config):
        self.encoder.set_sample_config(subnetwork_config)
        self.pooler.set_sample_config(subnetwork_config)


class BertSupernetForSequenceClassification(BertForSequenceClassification):
    def __init__(self, config, num_labels):
        super().__init__(config)
        self.bert = BertSupernetModel(config)
        self.classifier = LinearSuper(config.hidden_size, config.num_labels)

    def set_sample_config(self, subnetwork_config):
        self.bert.set_sample_config(subnetwork_config)
        self.subnetwork_hidden_size = subnetwork_config["hidden_size"]
        self.classifier.set_sample_config(self.subnetwork_hidden_size, self.num_labels)



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import math
from typing import List

import numpy as np
import pandas as pd
from scipy.spatial import Delaunay
from shapely.geometry import MultiLineString, MultiPoint, mapping
from shapely.ops import cascaded_union, polygonize
from matplotlib.cm import ScalarMappable
import matplotlib.pyplot as plt
# from dynast.supernetwork.supernetwork_registry import SUPERNET_METRICS

from dynast.utils import log


def frontier_builder(df, optimization_metrics, alpha=0, verbose=False):
    """
    Modified alphashape algorithm to draw Pareto Front for OFA search.
    Takes a DataFrame of column form [x, y] = [latency, accuracy]

    Params:
    df     - dataframe containing `optimization_metrics` columns at minimum
    alpha  - Dictates amount of tolerable 'concave-ness' allowed.
             A fully convex front will be given if 0 (also better for runtime)
    """
    if verbose:
        log.info('Running front builder')
    df = df[optimization_metrics]
    points = list(df.to_records(index=False))
    for i in range(len(points)):
        points[i] = list(points[i])
    points = MultiPoint(points)

    # TODO(macsz) Fix the line below in comment:
    # if len(points) < 4 or alpha <= 0:
    if alpha <= 0:
        if verbose:
            log.info('Alpha=0 -> convex hull')
        result = points.convex_hull
    else:
        coords = np.array([point.coords[0] for point in points])
        tri = Delaunay(coords)
        edges = set()
        edge_points = []
        edge_out = []

        # Loop over triangles
        for ia, ib, ic in tri.vertices:
            pa = coords[ia]
            pb = coords[ib]
            pc = coords[ic]

            # Lengths of sides of triangle
            a = math.sqrt((pa[0] - pb[0]) ** 2 + (pa[1] - pb[1]) ** 2)
            b = math.sqrt((pb[0] - pc[0]) ** 2 + (pb[1] - pc[1]) ** 2)
            c = math.sqrt((pc[0] - pa[0]) ** 2 + (pc[1] - pa[1]) ** 2)

            # Semiperimeter of triangle
            s = (a + b + c) * 0.5

            # Area of triangle by Heron's formula
            # Precompute value inside square root to avoid unbound math error in
            # case of 0 area triangles.
            area = s * (s - a) * (s - b) * (s - c)

            if area > 0:
                area = math.sqrt(area)

                # Radius Filter
                if a * b * c / (4.0 * area) < 1.0 / alpha:
                    for i, j in itertools.combinations([ia, ib, ic], r=2):
                        if (i, j) not in edges and (j, i) not in edges:
                            edges.add((i, j))
                            edge_points.append(coords[[i, j]])

                            if coords[i].tolist() not in edge_out:
                                edge_out.append(coords[i].tolist())
                            if coords[j].tolist() not in edge_out:
                                edge_out.append(coords[j].tolist())

        # Create the resulting polygon from the edge points
        m = MultiLineString(edge_points)
        triangles = list(polygonize(m))
        result = cascaded_union(triangles)

    # Find multi-polygon boundary
    bound = list(mapping(result.boundary)['coordinates'])

    # Cutoff non-Pareto front points
    # note that extreme concave geometries will create issues if bi-sected by line
    df = pd.DataFrame(bound, columns=['x', 'y'])

    # y=mx+b
    left_point = (df.iloc[df.idxmin()[0]][0], df.iloc[df.idxmin()[0]][1])
    right_point = (df.iloc[df.idxmax()[1]][0], df.iloc[df.idxmax()[1]][1])
    m = (left_point[1] - right_point[1]) / (left_point[0] - right_point[0])
    b = left_point[1] - (m * left_point[0])

    df = df[df['y'] >= (m * df['x'] + b)]
    df.sort_values(by='x', inplace=True)
    df.reset_index(drop=True, inplace=True)

    # Cleanup - insure accuracy is always increasing with latency up the Pareto front
    best_acc = 0
    drop_list = []
    for i in range(len(df)):
        if df.iloc[i]['y'] > best_acc:
            best_acc = df.iloc[i]['y']
        else:
            drop_list.append(i)
    df.drop(df.index[drop_list], inplace=True)
    df.reset_index(drop=True, inplace=True)

    df.columns = optimization_metrics

    return df


def plot_search_progression(
        csv_path: str,
        plot_columns: List[str] = ['macs', 'accuracy_top1'],
        supernet_metrics: List[str] = ['params', 'latency', 'macs', 'accuracy_top1'],
        save_path: str=None,
    ):
    assert len(plot_columns) == 2, 'Only 2D plots are supported'

    df_dynas = pd.read_csv(csv_path)
    df_dynas.columns = ['config', 'date'] + supernet_metrics

    fig, ax = plt.subplots(figsize=(7,5))

    cm = plt.cm.get_cmap('viridis_r')
    count = [x for x in range(len(df_dynas))]

    ax.scatter(df_dynas[plot_columns[0]].values, df_dynas[plot_columns[1]].values, marker='^', alpha=0.8, c=count,
            cmap=cm, label='Discovered DNN Model', s=10)

    df_front = frontier_builder(df_dynas, plot_columns)
    ax.plot(
        df_front[plot_columns[0]].values,
        df_front[plot_columns[1]].values,
        '--',
        color='red',
        label='Pareto Frontier',
        linewidth=2,
    )

    ax.set_title('DyNAS-T Search Results\n OFA ResNet50 - Bug')
    ax.set_xlabel(plot_columns[0], fontsize=13)
    ax.set_ylabel(plot_columns[1], fontsize=13)
    ax.legend(fancybox=True, fontsize=10, framealpha=1, borderpad=0.2, loc='lower right')
    ax.grid(True, alpha=0.3)
    #ax.set_ylim(72,77.5)

    # ax.plot(1.75e10, 77.9, marker='*', markersize=10, color='red', label='ViT Supernet')

    # Eval Count bar
    norm = plt.Normalize(0, len(df_dynas))
    sm = ScalarMappable(norm=norm, cmap=cm)
    cbar = fig.colorbar(sm, ax=ax, shrink=0.85)
    cbar.ax.set_title("         Evaluation\n  Count", fontsize=8)

    fig.tight_layout(pad=2)

    if save_path is None:
        save_path = csv_path.replace('.csv', '.png')
    log.info(f'Search progression visualization saved to {save_path}')
    plt.savefig(save_path, dpi=300)


def plot_search_progression_2(
        csv_path_1: str,
        csv_path_2: str,
        plot_columns: List[str] = ['macs', 'accuracy_top1'],
        supernet_metrics: List[str] = ['params', 'latency', 'macs', 'accuracy_top1'],
        save_path: str=None,
    ):
    assert len(plot_columns) == 2, 'Only 2D plots are supported'

    # 1
    df_dynas = pd.read_csv(csv_path_1)
    df_dynas.columns = ['config', 'date'] + supernet_metrics

    fig, ax = plt.subplots(figsize=(7,5))

    cm = plt.cm.get_cmap('Oranges')
    count = [x for x in range(len(df_dynas))]

    ax.scatter(df_dynas[plot_columns[0]].values, df_dynas[plot_columns[1]].values, marker='^', alpha=0.8, c=count,
            cmap=cm, label='Discovered DNN Model - Bug', s=10)

    df_front = frontier_builder(df_dynas, plot_columns)
    ax.plot(
        df_front[plot_columns[0]].values,
        df_front[plot_columns[1]].values,
        '--',
        color='tab:orange',
        label='Pareto Frontier - Bug',
        linewidth=2,
    )

    # 2
    df_dynas = pd.read_csv(csv_path_2)
    df_dynas.columns = ['config', 'date'] + supernet_metrics

    # fig, ax = plt.subplots(figsize=(7,5))

    cm = plt.cm.get_cmap('Greens')
    count = [x for x in range(len(df_dynas))]

    ax.scatter(df_dynas[plot_columns[0]].values, df_dynas[plot_columns[1]].values, marker='^', alpha=0.8, c=count,
            cmap=cm, label='Discovered DNN Model - Fixed', s=10)

    df_front = frontier_builder(df_dynas, plot_columns)
    ax.plot(
        df_front[plot_columns[0]].values,
        df_front[plot_columns[1]].values,
        '--',
        color='tab:green',
        label='Pareto Frontier - Fixed',
        linewidth=2,
    )

    ax.set_title('DyNAS-T Search Results\n OFA ResNet50 - Bug')
    ax.set_xlabel(plot_columns[0], fontsize=13)
    ax.set_ylabel(plot_columns[1], fontsize=13)
    ax.legend(fancybox=True, fontsize=10, framealpha=1, borderpad=0.2, loc='lower right')
    ax.grid(True, alpha=0.3)
    #ax.set_ylim(72,77.5)

    # ax.plot(1.75e10, 77.9, marker='*', markersize=10, color='red', label='ViT Supernet')

    # Eval Count bar
    norm = plt.Normalize(0, len(df_dynas))
    sm = ScalarMappable(norm=norm, cmap=plt.cm.get_cmap('Greys'))
    cbar = fig.colorbar(sm, ax=ax, shrink=0.85)
    cbar.ax.set_title("         Evaluation\n  Count", fontsize=8)

    fig.tight_layout(pad=2)

    if save_path is None:
        save_path = csv_path_1.replace('.csv', '_comp.png')
    log.info(f'Search progression visualization saved to {save_path}')
    plt.savefig(save_path, dpi=300)


if __name__ == '__main__':
    # df_dynas = pd.read_csv('/localdisk/maciej/code/DyNAS-T-validate/bert_random.csv')[:evals]
    # df_dynas = pd.read_csv('/localdisk/maciej/code/DyNAS-T-validate/ofa_mbv3_linas_fr0.2.csv')
    # df_dynas = pd.read_csv('/localdisk/maciej/code/DyNAS-T-validate/bert_linas_column_fix.csv')
    plot_search_progression(
        # 'bert_linas_column_fix_2.csv',
        # 'bert_linas_main.csv',
        'results_ofaresnet50_linas_long_main.csv',
        # 'results_ofaresnet50_linas_long_ksed_fix.csv',
        # 'ofa_mbv3_d234_e346_k357_w1.0',
        plot_columns=['macs', 'accuracy_top1']
    )

    plot_search_progression_2(
        # 'bert_linas_column_fix_2.csv',
        # 'bert_linas_main.csv',
        'results_ofaresnet50_linas_long_main.csv',
        'results_ofaresnet50_linas_long_ksed_fix.csv',
        # 'ofa_mbv3_d234_e346_k357_w1.0',
        plot_columns=['macs', 'accuracy_top1']
    )


# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from glob import glob

IGNORE_DIRS = [
    '.tox',
    'dist',
    '.git',
    'htmlcov',
    'log_dir',
    '.idea',
    '.venv',
    'build',
]


def get_python_files(root_dir, exclude_files=None):
    if not exclude_files:
        exclude_files = []

    files = []
    code_dirs = [d for d in os.listdir(root_dir) if os.path.isdir(d) and d not in IGNORE_DIRS]
    pattern = "*.py"

    for code_dir in code_dirs:
        for d, _, _ in os.walk(code_dir):
            files.extend(glob(os.path.join(d, pattern)))

    return list(sorted(set(files) - set(exclude_files)))



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

import autograd.numpy as anp
import numpy as np
from pymoo.algorithms.moo.age import AGEMOEA  # TODO(macsz) Add lazy imports
from pymoo.algorithms.moo.ctaea import CTAEA
from pymoo.algorithms.moo.moead import MOEAD
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.algorithms.moo.rnsga2 import RNSGA2
from pymoo.algorithms.moo.unsga3 import UNSGA3
from pymoo.algorithms.soo.nonconvex.cmaes import CMAES
from pymoo.algorithms.soo.nonconvex.ga import GA
from pymoo.core.problem import Problem
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.repair.rounding import RoundingRepair
from pymoo.operators.sampling.rnd import IntegerRandomSampling
from pymoo.optimize import minimize
from pymoo.util.ref_dirs import get_reference_directions

from dynast.utils import log


class EvolutionaryManager:
    '''
    Manages the search parameters for the DyNAS-T single/multi-objective search.

    Parameters
    ----------
    algorithm : string
        Define a multi-objective search algorithm.
    seed : int
        Seed value for Pymoo search.
    verbose : Boolean
        Verbosity option
    engine : string
        Support different engine types (e.g. pymoo)
    n_obj: int
        Number of objectives to optimize for.
    '''

    def __init__(
        self,
        algorithm: str = 'nsga2',
        seed: int = 0,
        verbose: bool = False,
        engine: str = 'pymoo',
        n_obj: int = 2,
    ):
        self.algorithm = algorithm
        self.seed = seed
        self.verbose = verbose
        self.engine = engine
        self.n_obj = n_obj
        self.set_algorithm_def()

    def set_algorithm_def(self):
        if self.algorithm == 'cmaes' and self.n_obj == 1:
            self.algorithm_def = self.configure_cmaes()
            self.engine = 'pymoo'
        elif self.algorithm == 'ga' and self.n_obj == 1:
            self.algorithm_def = self.configure_ga()
            self.engine = 'pymoo'
        elif self.algorithm == 'nsga2' and self.n_obj == 2:
            self.algorithm_def = self.configure_nsga2()
            self.engine = 'pymoo'
        elif self.algorithm == 'rnsga2' and self.n_obj == 2:
            self.algorithm_def = self.configure_rnsga2()
            self.engine = 'pymoo'
        elif self.algorithm == 'age' and self.n_obj == 2:
            self.algorithm_def = self.configure_age()
            self.engine = 'pymoo'
        elif self.algorithm == 'ctaea' and self.n_obj == 3:
            self.algorithm_def = self.configure_ctaea()
            self.engine = 'pymoo'
        elif self.algorithm == 'moead' and self.n_obj == 3:
            self.algorithm_def = self.configure_moead()
            self.engine = 'pymoo'
        elif self.algorithm == 'unsga3' and self.n_obj == 3:
            self.algorithm_def = self.configure_unsga3()
            self.engine = 'pymoo'
        else:
            log.warning(f'Algorithm {self.algorithm} and/or number of objectives {self.n_obj} not supported.')
            raise NotImplementedError

        log.info(f'Configuring {self.algorithm} algorithm with {self.n_obj} objectives.')

    def configure_cmaes(
        self,
        num_evals: int = 1000,
        restarts: int = 10,
        restart_from_best: bool = True,
        warm_pop: np.ndarray = None,
    ):
        '''
        Refer to: https://pymoo.org/algorithms/soo/cmaes.html
        '''

        # CMA does not have "populations" in the genetic sense. Thus n_gens is set to number of evals.
        self.n_gens = num_evals

        if warm_pop:
            log.info('Using warm start population.')
            sample_strategy = warm_pop[0]  # TODO(macsz) This is not used - drop?
        else:
            sample_strategy = None

        self.algorithm_def = CMAES(
            restarts=10,
            restart_from_best=restart_from_best,
            maxfevals=int(self.n_gens) * restarts,
            x0=None,
        )

    def configure_ga(
        self,
        population=50,
        num_evals=1000,
        warm_pop=None,
        crossover_prob=0.9,
        crossover_eta=15.0,
        mutation_prob=0.02,
        mutation_eta=20.0,
    ):
        '''
        Simple Genetic Algorithm https://pymoo.org/algorithms/soo/ga.html
        '''

        self.n_gens = num_evals / population

        if type(warm_pop) == 'numpy.ndarray':
            log.info('Using warm start population.')
            sample_strategy = warm_pop
        else:
            sample_strategy = IntegerRandomSampling()

        self.algorithm_def = GA(
            pop_size=population,
            sampling=sample_strategy,
            crossover=SBX(prob=crossover_prob, eta=crossover_eta, vtype=float, repair=RoundingRepair()),
            mutation=PM(prob=mutation_prob, eta=mutation_eta, vtype=float, repair=RoundingRepair()),
            eliminate_duplicates=True,
        )

    def configure_nsga2(
        self,
        population=50,
        num_evals=1000,
        warm_pop=None,
        crossover_prob=0.9,
        crossover_eta=15.0,
        mutation_prob=0.02,
        mutation_eta=20.0,
    ):
        self.n_gens = num_evals / population

        if type(warm_pop) == 'numpy.ndarray':
            log.info('Using warm start population.')
            sample_strategy = warm_pop
        else:
            sample_strategy = IntegerRandomSampling()

        self.algorithm_def = NSGA2(
            pop_size=population,
            sampling=sample_strategy,
            crossover=SBX(prob=crossover_prob, eta=crossover_eta, vtype=float, repair=RoundingRepair()),
            mutation=PM(prob=mutation_prob, eta=mutation_eta, vtype=float, repair=RoundingRepair()),
            eliminate_duplicates=True,
        )

    def configure_unsga3(
        self,
        population=50,
        num_evals=1000,
        ref_dirs=None,
        warm_pop=None,
        crossover_prob=1.0,
        crossover_eta=30.0,
        mutation_prob=0.02,
        mutation_eta=20.0,
        n_partitions=20,
    ):
        self.n_gens = num_evals / population

        ref_dirs = get_reference_directions("energy", self.n_obj, n_partitions, seed=0)
        ref_dirs = ref_dirs.astype('float64')

        if type(warm_pop) == 'numpy.ndarray':
            log.info('Using warm start population')
            sample_strategy = warm_pop
        else:
            sample_strategy = IntegerRandomSampling()

        self.algorithm_def = UNSGA3(
            ref_dirs=ref_dirs,
            pop_size=population,
            sampling=sample_strategy,
            crossover=SBX(prob=crossover_prob, eta=crossover_eta, vtype=float, repair=RoundingRepair()),
            mutation=PM(prob=mutation_prob, eta=mutation_eta, vtype=float, repair=RoundingRepair()),
            eliminate_duplicates=True,
        )

    def configure_rnsga2(
        self,
        population=50,
        num_evals=1000,
        warm_pop=None,
        ref_points=[[0, 0]],
        crossover_prob=0.9,
        crossover_eta=15.0,
        mutation_prob=0.02,
        mutation_eta=20.0,
    ):
        self.engine = 'pymoo'
        self.n_gens = num_evals / population

        if type(warm_pop) == 'numpy.ndarray':
            log.info('Using warm start population')
            sample_strategy = warm_pop
        else:
            sample_strategy = IntegerRandomSampling()

        log.info('Reference points for RNSGA-II are: {}'.format(ref_points))
        reference_points = np.array(ref_points)  # lat=0, 1/acc=0

        self.algorithm_def = RNSGA2(
            ref_points=reference_points,
            pop_size=population,
            epsilon=0.01,
            normalization='front',
            extreme_points_as_reference_points=False,
            sampling=sample_strategy,
            crossover=SBX(prob=crossover_prob, eta=crossover_eta, vtype=float, repair=RoundingRepair()),
            mutation=PM(prob=mutation_prob, eta=mutation_eta, vtype=float, repair=RoundingRepair()),
            weights=np.array([0.5, 0.5]),
            eliminate_duplicates=True,
        )

    def configure_age(
        self,
        population=50,
        num_evals=1000,
        warm_pop=None,
        crossover_prob=0.9,
        crossover_eta=15.0,
        mutation_prob=0.02,
        mutation_eta=20.0,
    ):
        self.engine = 'pymoo'
        self.n_gens = num_evals / population

        if type(warm_pop) == 'numpy.ndarray':
            log.info('[Info] Using warm start population')
            sample_strategy = warm_pop
        else:
            sample_strategy = IntegerRandomSampling()

        self.algorithm_def = AGEMOEA(
            pop_size=population,
            sampling=sample_strategy,
            crossover=SBX(prob=crossover_prob, eta=crossover_eta, vtype=float, repair=RoundingRepair()),
            mutation=PM(prob=mutation_prob, eta=mutation_eta, vtype=float, repair=RoundingRepair()),
            eliminate_duplicates=True,
        )

    def configure_ctaea(self, warm_pop=None, num_evals=1000, ref_dirs=None, n_partitions=20):
        self.engine = 'pymoo'
        self.n_gens = num_evals / n_partitions

        if type(warm_pop) == 'numpy.ndarray':
            log.info('Using warm start population')
            sample_strategy = warm_pop
        else:
            sample_strategy = IntegerRandomSampling()

        ref_dirs = get_reference_directions("energy", self.n_obj, n_partitions, seed=0)
        ref_dirs = ref_dirs.astype('float64')

        self.algorithm_def = CTAEA(ref_dirs=ref_dirs, sampling=sample_strategy, eliminate_duplicates=True)

    def configure_moead(
        self,
        n_neighbors=20,
        num_evals=1000,
        warm_pop=None,
        ref_dirs=None,
        crossover_prob=1.0,
        crossover_eta=20.0,
        mutation_prob=None,
        mutation_eta=20.0,
        n_partitions=20,
    ):
        self.engine = 'pymoo'
        self.n_gens = num_evals / n_neighbors

        log.info('Configuring MOEA/D algorithm.')

        if type(warm_pop) == 'numpy.ndarray':
            log.info('Using warm start population')
            sample_strategy = warm_pop
        else:
            sample_strategy = IntegerRandomSampling()

        ref_dirs = get_reference_directions("energy", self.n_obj, n_partitions, seed=self.seed)
        ref_dirs = ref_dirs.astype('float64')

        self.algorithm_def = MOEAD(
            ref_dirs=ref_dirs,
            n_neighbors=n_neighbors,
            sampling=sample_strategy,
            crossover=SBX(prob=crossover_prob, eta=crossover_eta, vtype=float, repair=RoundingRepair()),
            mutation=PM(prob=mutation_prob, eta=mutation_eta, vtype=float, repair=RoundingRepair()),
        )

    def run_search(self, problem, save_history=False):
        '''
        Note: Known issue (memory leak) with the pymoo save history option v0.5.0
        '''

        log.info('Running Search ...')
        start_time = time.time()

        if self.engine == 'pymoo':
            result = minimize(
                problem,
                self.algorithm_def,
                ('n_gen', int(self.n_gens)),
                seed=int(self.seed),
                save_history=save_history,
                verbose=self.verbose,
            )
        else:
            log.info('Invalid algorithm engine configuration!')
            raise NotImplementedError

        log.info('Success! Search Took {:.3f} seconds.'.format(time.time() - start_time))

        return result


class EvolutionarySingleObjective(Problem):
    '''
    Interface between the user-defined evaluation interface and the SearchAlgoManager.

    Parameters
    ----------
    evaluation_interface : Class
        Class that handles the objective measurement call from the supernet.
    param_count : int
        Number variables in the search space (e.g., OFA MobileNetV3 has 45)
    param_upperbound : array
        The upper int array that defines how many options each design variable has.
    '''

    def __init__(self, evaluation_interface, param_count, param_upperbound):
        super().__init__(n_var=param_count, n_obj=1, n_constr=0, xl=0, xu=param_upperbound, type_var=int)

        self.evaluation_interface = evaluation_interface

    def _evaluate(self, x, out, *args, **kwargs):
        # Store results for a given generation for PyMoo
        objective_arr = list()

        # Measure new individuals
        for i in range(len(x)):
            _, objective = self.evaluation_interface.eval_subnet(x[i])

            objective_arr.append(objective)

        print('.', end='', flush=True)

        # Update PyMoo with evaluation data
        out["F"] = anp.column_stack([objective_arr])


class EvolutionaryMultiObjective(Problem):
    '''
    Interface between the user-defined evaluation interface and the SearchAlgoManager.

    Parameters
    ----------
    evaluation_interface : Class
        Class that handles the objective measurement call from the supernet.
    param_count : int
        Number variables in the search space (e.g., OFA MobileNetV3 has 45)
    param_upperbound : array
        The upper int array that defines how many options each design variable has.
    '''

    def __init__(self, evaluation_interface, param_count, param_upperbound):
        super().__init__(n_var=param_count, n_obj=2, n_constr=0, xl=0, xu=param_upperbound, type_var=int)

        self.evaluation_interface = evaluation_interface

    def _evaluate(self, x, out, *args, **kwargs):
        # Store results for a given generation for PyMoo
        objective_x_arr, objective_y_arr = list(), list()

        # Measure new individuals
        for i in range(len(x)):
            _, objective_x, objective_y = self.evaluation_interface.eval_subnet(x[i])

            objective_x_arr.append(objective_x)
            objective_y_arr.append(objective_y)

        print('.', end='', flush=True)

        # Update PyMoo with evaluation data
        out["F"] = anp.column_stack([objective_x_arr, objective_y_arr])


class EvolutionaryManyObjective(Problem):
    '''
    Interface between the user-defined evaluation interface and the SearchAlgoManager.

    Parameters
    ----------
    evaluation_interface : Class
        Class that handles the objective measurement call from the supernet.
    param_count : int
        Number variables in the search space (e.g., OFA MobileNetV3 has 45)
    param_upperbound : array
        The upper int array that defines how many options each design variable has.
    '''

    def __init__(self, evaluation_interface, param_count, param_upperbound):
        super().__init__(n_var=param_count, n_obj=3, n_constr=0, xl=0, xu=param_upperbound, type_var=int)

        self.evaluation_interface = evaluation_interface

    def _evaluate(self, x, out, *args, **kwargs):
        # Store results for a given generation for PyMoo
        objective_x_arr, objective_y_arr, objective_z_arr = list(), list(), list()

        # Measure new individuals
        for i in range(len(x)):
            _, objective_x, objective_y, objective_z = self.evaluation_interface.eval_subnet(x[i])

            objective_x_arr.append(objective_x)
            objective_y_arr.append(objective_y)
            objective_z_arr.append(objective_z)

        print('.', end='', flush=True)

        # Update PyMoo with evaluation data
        out["F"] = anp.column_stack([objective_x_arr, objective_y_arr, objective_z_arr])



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict

import torch
import torch.nn as nn
import torch.nn.functional as F
from fairseq import utils
from torch.nn import Parameter
from torch.nn.modules.module import _addindent

INCREMENTAL_STATE_INSTANCE_ID = defaultdict(lambda: 0)


def _get_full_incremental_state_key(module_instance, key):
    module_name = module_instance.__class__.__name__

    # assign a unique ID to each module instance, so that incremental state is
    # not shared across module instances
    if not hasattr(module_instance, '_fairseq_instance_id'):
        INCREMENTAL_STATE_INSTANCE_ID[module_name] += 1
        module_instance._fairseq_instance_id = INCREMENTAL_STATE_INSTANCE_ID[module_name]

    return '{}.{}.{}'.format(module_name, module_instance._fairseq_instance_id, key)


def get_incremental_state(module, incremental_state, key):
    """Helper for getting incremental state for an nn.Module."""
    full_key = _get_full_incremental_state_key(module, key)
    if incremental_state is None or full_key not in incremental_state:
        return None
    return incremental_state[full_key]


def set_incremental_state(module, incremental_state, key, value):
    """Helper for setting incremental state for an nn.Module."""
    if incremental_state is not None:
        full_key = _get_full_incremental_state_key(module, key)
        incremental_state[full_key] = value


class EmbeddingSuper(nn.Embedding):
    def __init__(self, num_embeddings, super_embed_dim, padding_idx, *args, **kwargs):
        super().__init__(num_embeddings, super_embed_dim, padding_idx, *args, **kwargs)

        # the largest embed dim
        self.super_embed_dim = {'encoder': super_embed_dim, 'decoder': super_embed_dim}

        # the current sampled embed dim
        self.sample_embed_dim = {'encoder': None, 'decoder': None}

        self.samples = {'encoder': {}, 'decoder': {}}
        self.profiling = False
        self.reset_parameters()

    def profile(self, mode=True):
        self.profiling = mode

    def reset_parameters(self):
        super().reset_parameters()
        nn.init.normal_(self.weight, mean=0, std=self.embedding_dim**-0.5)
        nn.init.constant_(self.weight[self.padding_idx], 0)

    def set_sample_config(self, sample_embed_dim, part):
        self.sample_embed_dim[part] = sample_embed_dim
        self._sample_parameters(part)

    def _sample_parameters(self, part):
        weight = self.weight[..., : self.sample_embed_dim[part]]
        self.samples[part]['weight'] = weight

        return self.samples

    def sample_parameters(self, part, resample=False):
        return self._sample_parameters(part) if self.profiling or resample else self.samples

    def sampled_weight(self, part):
        return self.sample_parameters(part)[part]['weight']

    def forward(self, input, part='encoder'):
        return F.embedding(
            input,
            self.sampled_weight(part),
            self.padding_idx,
            self.max_norm,
            self.norm_type,
            self.scale_grad_by_freq,
            self.sparse,
        )


class LinearSuper(nn.Linear):
    def __init__(self, super_in_dim, super_out_dim, bias=True, uniform_=None, non_linear='linear'):
        super().__init__(super_in_dim, super_out_dim, bias=bias)

        # super_in_dim and super_out_dim indicate the largest network!
        self.super_in_dim = super_in_dim
        self.super_out_dim = super_out_dim

        # input_dim and output_dim indicate the current sampled size
        self.sample_in_dim = None
        self.sample_out_dim = None

        self.samples = {}

        self._reset_parameters(bias, uniform_, non_linear)
        self.profiling = False

    def profile(self, mode=True):
        self.profiling = mode

    def sample_parameters(self, resample=False):
        if self.profiling or resample:
            return self._sample_parameters()
        return self.samples

    def _reset_parameters(self, bias, uniform_, non_linear):
        nn.init.xavier_uniform_(self.weight) if uniform_ is None else uniform_(self.weight, non_linear=non_linear)
        if bias:
            nn.init.constant_(self.bias, 0.0)

    def set_sample_config(self, sample_in_dim, sample_out_dim):
        self.sample_in_dim = sample_in_dim
        self.sample_out_dim = sample_out_dim

        self._sample_parameters()

    def _sample_parameters(self):
        self.samples['weight'] = sample_weight(self.weight, self.sample_in_dim, self.sample_out_dim)
        self.samples['bias'] = self.bias
        if self.bias is not None:
            self.samples['bias'] = sample_bias(self.bias, self.sample_out_dim)
        return self.samples

    def forward(self, x):
        self.sample_parameters()
        return F.linear(x, self.samples['weight'], self.samples['bias'])

    def calc_sampled_param_num(self):
        assert 'weight' in self.samples.keys()
        weight_numel = self.samples['weight'].numel()

        if self.samples['bias'] is not None:
            bias_numel = self.samples['bias'].numel()
        else:
            bias_numel = 0

        return weight_numel + bias_numel


def sample_weight(weight, sample_in_dim, sample_out_dim):
    sample_weight = weight[:, :sample_in_dim]
    sample_weight = sample_weight[:sample_out_dim, :]

    return sample_weight


def sample_bias(bias, sample_out_dim):
    sample_bias = bias[:sample_out_dim]

    return sample_bias


def LayerNorm(normalized_shape, eps=1e-5, elementwise_affine=True, export=False):
    if not export and torch.cuda.is_available():
        try:
            from apex.normalization import FusedLayerNorm

            return FusedLayerNorm(normalized_shape, eps, elementwise_affine)
        except ImportError:
            pass
    return torch.nn.LayerNorm(normalized_shape, eps, elementwise_affine)


class LayerNormSuper(torch.nn.LayerNorm):
    def __init__(self, super_embed_dim):
        super().__init__(super_embed_dim)

        # the largest embed dim
        self.super_embed_dim = super_embed_dim

        # the current sampled embed dim
        self.sample_embed_dim = None

        self.samples = {}
        self.profiling = False

    def profile(self, mode=True):
        self.profiling = mode

    def sample_parameters(self, resample=False):
        if self.profiling or resample:
            return self._sample_parameters()
        return self.samples

    def _sample_parameters(self):
        self.samples['weight'] = self.weight[: self.sample_embed_dim]
        self.samples['bias'] = self.bias[: self.sample_embed_dim]
        return self.samples

    def set_sample_config(self, sample_embed_dim):
        self.sample_embed_dim = sample_embed_dim
        self._sample_parameters()

    def forward(self, x):
        self.sample_parameters()
        return F.layer_norm(
            x, (self.sample_embed_dim,), weight=self.samples['weight'], bias=self.samples['bias'], eps=self.eps
        )

    def calc_sampled_param_num(self):
        assert 'weight' in self.samples.keys()
        assert 'bias' in self.samples.keys()
        return self.samples['weight'].numel() + self.samples['bias'].numel()


class MultiheadAttentionSuper(nn.Module):
    """Multi-headed attention.

    See "Attention Is All You Need" for more details.
    """

    def __init__(
        self,
        super_embed_dim,
        num_heads,
        is_encoder,
        super_kdim=None,
        super_vdim=None,
        dropout=0.0,
        bias=True,
        add_bias_kv=False,
        add_zero_attn=False,
        self_attention=False,
        encoder_decoder_attention=False,
        out_dim=None,
        qkv_dim=None,
    ):
        super().__init__()

        # the configs of super arch
        self.super_q_embed_dim = super_embed_dim
        self.super_kv_embed_dim = None

        # the configs of current sampled arch
        self.sample_q_embed_dim = None
        self.sample_kv_embed_dim = None

        if super_kdim is not None:
            assert super_kdim == super_vdim
            self.super_kv_embed_dim = super_kdim
        else:
            self.super_kv_embed_dim = self.super_q_embed_dim

        if qkv_dim is None:
            self.qkv_dim = self.super_q_embed_dim
        else:
            self.qkv_dim = qkv_dim

        # this qkv same dim means the input dim for qkv are the same, not the output dim
        # self.qkv_same_dim = self.kdim == self.super_embed_dim and self.vdim == self.super_embed_dim
        self.qkv_same_dim = self.super_kv_embed_dim == self.super_q_embed_dim
        self.encoder = is_encoder

        # Caution! these actually are the sampled num_heads, head_dim and scaling
        self.num_heads = num_heads
        self.dropout = dropout
        self.head_dim = self.qkv_dim // num_heads
        assert self.head_dim * num_heads == self.qkv_dim, "qkv must be divisible by num_heads"
        self.scaling = self.head_dim**-0.5

        self.self_attention = self_attention
        self.encoder_decoder_attention = encoder_decoder_attention

        assert not self.self_attention or self.qkv_same_dim, (
            'Self-attention requires query, key and ' 'value to be of the same size'
        )

        if self.qkv_same_dim:
            self.in_proj_weight = Parameter(torch.Tensor(3 * self.qkv_dim, self.super_q_embed_dim))
        else:
            self.k_proj_weight = Parameter(torch.Tensor(self.qkv_dim, self.super_kv_embed_dim))
            self.v_proj_weight = Parameter(torch.Tensor(self.qkv_dim, self.super_kv_embed_dim))
            self.q_proj_weight = Parameter(torch.Tensor(self.qkv_dim, self.super_q_embed_dim))

        if bias:
            self.in_proj_bias = Parameter(torch.Tensor(3 * self.qkv_dim))
        else:
            self.register_parameter('in_proj_bias', None)

        if out_dim is None:
            out_dim = self.super_q_embed_dim
        self.out_proj = LinearSuper(super_in_dim=self.qkv_dim, super_out_dim=out_dim, bias=bias)

        if add_bias_kv:
            self.bias_k = Parameter(torch.Tensor(1, 1, self.super_q_embed_dim))
            self.bias_v = Parameter(torch.Tensor(1, 1, self.super_q_embed_dim))
        else:
            self.bias_k = self.bias_v = None

        self.add_zero_attn = add_zero_attn

        self.reset_parameters()

        self.onnx_trace = False

        self.enable_torch_version = False
        if hasattr(F, "multi_head_attention_forward"):
            self.enable_torch_version = True
        else:
            self.enable_torch_version = False
        self.enable_torch_version = False

    def calc_sampled_param_num(self):
        assert self.in_proj_weight is not None and self.in_proj_bias is not None
        in_proj_q_weight_numel = self.sample_q_embed_dim * self.qkv_dim
        in_proj_v_weight_numel = in_proj_k_weight_numel = self.sample_kv_embed_dim * self.qkv_dim
        in_proj_bias_numel = self.in_proj_bias.numel()

        # does not count in the output proj because it will be counted in LinearSuper layer
        # out_proj_weight_numel = self.qkv_dim * self.sample_q_embed_dim
        # out_proj_bias_numel = self.

        return in_proj_q_weight_numel + in_proj_k_weight_numel + in_proj_v_weight_numel + in_proj_bias_numel

    def set_sample_config(self, sample_q_embed_dim, sample_attention_heads, sample_kv_embed_dim=None):
        self.sample_q_embed_dim = sample_q_embed_dim
        if sample_kv_embed_dim is None:
            self.sample_kv_embed_dim = sample_q_embed_dim
        else:
            self.sample_kv_embed_dim = sample_kv_embed_dim

        self.num_heads = sample_attention_heads
        self.head_dim = self.qkv_dim // self.num_heads
        assert self.head_dim * self.num_heads == self.qkv_dim, "qkv_dim must be divisible by sampled num_heads"
        self.scaling = self.head_dim**-0.5

        self.out_proj.set_sample_config(sample_in_dim=self.qkv_dim, sample_out_dim=self.sample_q_embed_dim)

    def prepare_for_onnx_export_(self):
        self.onnx_trace = True

    def reset_parameters(self):
        if self.qkv_same_dim:
            nn.init.xavier_uniform_(self.in_proj_weight)
        else:
            nn.init.xavier_uniform_(self.k_proj_weight)
            nn.init.xavier_uniform_(self.v_proj_weight)
            nn.init.xavier_uniform_(self.q_proj_weight)

        nn.init.xavier_uniform_(self.out_proj.weight)
        if self.in_proj_bias is not None:
            nn.init.constant_(self.in_proj_bias, 0.0)
            nn.init.constant_(self.out_proj.bias, 0.0)
        if self.bias_k is not None:
            nn.init.xavier_normal_(self.bias_k)
        if self.bias_v is not None:
            nn.init.xavier_normal_(self.bias_v)

    def forward(
        self,
        query,
        key,
        value,
        key_padding_mask=None,
        incremental_state=None,
        need_weights=True,
        static_kv=False,
        attn_mask=None,
    ):
        """Input shape: Time x Batch x Channel

        Timesteps can be masked by supplying a T x T mask in the
        `attn_mask` argument. Padding elements can be excluded from
        the key by passing a binary ByteTensor (`key_padding_mask`) with shape:
        batch x src_len, where padding elements are indicated by 1s.
        """

        tgt_len, bsz, embed_dim = query.size()

        if incremental_state is not None:
            saved_state = self._get_input_buffer(incremental_state)
            if 'prev_key' in saved_state:
                # previous time steps are cached - no need to recompute
                # key and value if they are static
                if static_kv:
                    assert self.encoder_decoder_attention and not self.self_attention
                    key = value = None
        else:
            saved_state = None

        if self.self_attention:
            # self-attention
            q, k, v = self.in_proj_qkv(query)
        elif self.encoder_decoder_attention:
            # encoder-decoder attention
            q = self.in_proj_q(query)
            if key is None:
                assert value is None
                k = v = None
            else:
                k = self.in_proj_k(key)
                v = self.in_proj_v(key)

        else:
            q = self.in_proj_q(query)
            k = self.in_proj_k(key)
            v = self.in_proj_v(value)

        q = q * self.scaling

        if self.bias_k is not None:
            assert self.bias_v is not None
            k = torch.cat([k, self.bias_k.repeat(1, bsz, 1)])
            v = torch.cat([v, self.bias_v.repeat(1, bsz, 1)])
            if attn_mask is not None:
                attn_mask = torch.cat([attn_mask, attn_mask.new_zeros(attn_mask.size(0), 1)], dim=1)
            if key_padding_mask is not None:
                key_padding_mask = torch.cat(
                    [key_padding_mask, key_padding_mask.new_zeros(key_padding_mask.size(0), 1)], dim=1
                )

        q = q.contiguous().view(tgt_len, bsz * self.num_heads, self.head_dim).transpose(0, 1)

        if k is not None:
            k = k.contiguous().view(-1, bsz * self.num_heads, self.head_dim).transpose(0, 1)
        if v is not None:
            v = v.contiguous().view(-1, bsz * self.num_heads, self.head_dim).transpose(0, 1)

        if saved_state is not None:
            # saved states are stored with shape (bsz, num_heads, seq_len, head_dim)
            if 'prev_key' in saved_state:
                prev_key = saved_state['prev_key'].view(bsz * self.num_heads, -1, self.head_dim)
                if static_kv:
                    k = prev_key
                else:
                    k = torch.cat((prev_key, k), dim=1)
            if 'prev_value' in saved_state:
                prev_value = saved_state['prev_value'].view(bsz * self.num_heads, -1, self.head_dim)
                if static_kv:
                    v = prev_value
                else:
                    v = torch.cat((prev_value, v), dim=1)
            saved_state['prev_key'] = k.view(bsz, self.num_heads, -1, self.head_dim)
            if v is not None:
                # TODO(macsz) Static code analysis tool indicates variable `v` can be unitialized here.
                # However, I do not see how it's possible with current flow. Just in case raising
                # an error to indicate this code can be problematic.
                saved_state['prev_value'] = v.view(bsz, self.num_heads, -1, self.head_dim)
            else:
                raise UnboundLocalError

            self._set_input_buffer(incremental_state, saved_state)

        src_len = k.size(1)

        # This is part of a workaround to get around fork/join parallelism
        # not supporting Optional types.
        if key_padding_mask is not None and key_padding_mask.shape == torch.Size([]):
            key_padding_mask = None

        if key_padding_mask is not None:
            fil = key_padding_mask.new_ones(key_padding_mask.size(0), src_len - key_padding_mask.size(1))
            key_padding_mask = torch.cat((key_padding_mask, fil), dim=1)
            assert key_padding_mask.size(0) == bsz
            assert key_padding_mask.size(1) == src_len

        if self.add_zero_attn:
            src_len += 1
            k = torch.cat([k, k.new_zeros((k.size(0), 1) + k.size()[2:])], dim=1)
            if v is not None:
                # TODO(macsz) Static code analysis tool indicates variable `v` can be unitialized here.
                # However, I do not see how it's possible with current flow. Just in case raising
                # an error to indicate this code can be problematic.
                v = torch.cat([v, v.new_zeros((v.size(0), 1) + v.size()[2:])], dim=1)
            else:
                raise UnboundLocalError
            if attn_mask is not None:
                attn_mask = torch.cat([attn_mask, attn_mask.new_zeros(attn_mask.size(0), 1)], dim=1)
            if key_padding_mask is not None:
                key_padding_mask = torch.cat(
                    [key_padding_mask, torch.zeros(key_padding_mask.size(0), 1).type_as(key_padding_mask)], dim=1
                )

        attn_weights = torch.bmm(q, k.transpose(1, 2))
        attn_weights = self.apply_sparse_mask(attn_weights, tgt_len, src_len, bsz)

        assert list(attn_weights.size()) == [bsz * self.num_heads, tgt_len, src_len]

        if attn_mask is not None:
            attn_mask = attn_mask.unsqueeze(0)
            if self.onnx_trace:
                attn_mask = attn_mask.repeat(attn_weights.size(0), 1, 1)
            attn_weights += attn_mask

        if key_padding_mask is not None:
            attn_weights = attn_weights.view(bsz, self.num_heads, tgt_len, src_len)
            if self.onnx_trace:
                attn_weights = torch.where(
                    key_padding_mask.unsqueeze(1).unsqueeze(2), torch.Tensor([float("-Inf")]), attn_weights.float()
                ).type_as(attn_weights)
            else:
                attn_weights = attn_weights.masked_fill(
                    key_padding_mask.unsqueeze(1).unsqueeze(2),
                    float('-inf'),
                )
            attn_weights = attn_weights.view(bsz * self.num_heads, tgt_len, src_len)

        attn_weights = utils.softmax(
            attn_weights,
            dim=-1,
            onnx_trace=self.onnx_trace,
        ).type_as(attn_weights)
        attn_weights = F.dropout(attn_weights, p=self.dropout, training=self.training)

        attn = torch.bmm(attn_weights, v)

        assert list(attn.size()) == [bsz * self.num_heads, tgt_len, self.head_dim]

        if self.onnx_trace and attn.size(1) == 1:
            # when ONNX tracing a single decoder step (sequence length == 1)
            # the transpose is a no-op copy before view, thus unnecessary
            attn = attn.contiguous().view(tgt_len, bsz, self.qkv_dim)
        else:
            attn = attn.transpose(0, 1).contiguous().view(tgt_len, bsz, self.qkv_dim)
        attn = self.out_proj(attn)

        if need_weights:
            # average attention weights over heads
            attn_weights = attn_weights.view(bsz, self.num_heads, tgt_len, src_len)

            attn_weights = attn_weights.sum(dim=1) / self.num_heads
        else:
            attn_weights = None

        return attn, attn_weights

    def in_proj_qkv(self, query):
        return self._in_proj(query, sample_dim=self.sample_q_embed_dim).chunk(3, dim=-1)

    def in_proj_q(self, query):
        if self.qkv_same_dim:
            return self._in_proj(query, end=self.qkv_dim, sample_dim=self.sample_q_embed_dim)
        else:
            bias = self.in_proj_bias
            if bias is not None:
                bias = bias[: self.qkv_dim]
            return F.linear(query, self.q_proj_weight[..., : self.sample_q_embed_dim], bias)

    def in_proj_k(self, key):
        if self.qkv_same_dim:
            return self._in_proj(key, start=self.qkv_dim, end=2 * self.qkv_dim, sample_dim=self.sample_kv_embed_dim)
        else:
            weight = self.k_proj_weight
            bias = self.in_proj_bias
            if bias is not None:
                bias = bias[self.qkv_dim : 2 * self.qkv_dim]
            return F.linear(key, weight[..., : self.sample_kv_embed_dim], bias)

    def in_proj_v(self, value):
        if self.qkv_same_dim:
            return self._in_proj(value, start=2 * self.qkv_dim, sample_dim=self.sample_kv_embed_dim)
        else:
            weight = self.v_proj_weight
            bias = self.in_proj_bias
            if bias is not None:
                bias = bias[2 * self.qkv_dim :]
            return F.linear(value, weight[..., : self.sample_kv_embed_dim], bias)

    def _in_proj(self, input, sample_dim, start=0, end=None):
        weight = self.in_proj_weight
        bias = self.in_proj_bias
        weight = weight[start:end, :sample_dim]
        if bias is not None:
            bias = bias[start:end]
        return F.linear(input, weight, bias)

    def reorder_incremental_state(self, incremental_state, new_order):
        """Reorder buffered internal state (for incremental generation)."""
        input_buffer = self._get_input_buffer(incremental_state)
        if input_buffer is not None:
            for k in input_buffer.keys():
                input_buffer[k] = input_buffer[k].index_select(0, new_order)
            self._set_input_buffer(incremental_state, input_buffer)

    def _get_input_buffer(self, incremental_state):
        return (
            get_incremental_state(  # utils.
                self,
                incremental_state,
                'attn_state',
            )
            or {}
        )

    def _set_input_buffer(self, incremental_state, buffer):
        set_incremental_state(
            self,
            incremental_state,
            'attn_state',
            buffer,
        )

    def apply_sparse_mask(self, attn_weights, tgt_len, src_len, bsz):
        return attn_weights

    def __repr__(self):
        # We treat the extra repr like the sub-module, one item per line
        extra_lines = []
        extra_repr = self.extra_repr()
        # empty string will be split into list ['']
        if extra_repr:
            extra_lines = extra_repr.split('\n')
        child_lines = []
        for key, module in self._modules.items():
            mod_str = repr(module)
            mod_str = _addindent(mod_str, 2)
            child_lines.append('(' + key + '): ' + mod_str)
        lines = extra_lines + child_lines

        main_str = self._get_name() + '\tnum_heads:' + str(self.num_heads) + '\t qkv_dim:' + str(self.qkv_dim)
        if lines:
            # simple one-liner info, which most builtin Modules will use
            if len(extra_lines) == 1 and not child_lines:
                main_str += extra_lines[0]
            else:
                main_str += '\n  ' + '\n  '.join(lines) + '\n'

        main_str += ')'
        return main_str



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from unittest import mock
from unittest.mock import mock_open, patch

import pytest
import requests

from dynast.utils import check_kwargs_deprecated, get_cores, get_remote_file, split_list

valid_remote_url = 'http://someurl.com/test.txt'
valid_remote_url_file_not_exists = 'http://example.com/supernets/not_exists.txt'
existing_model_dir = '/tmp/'
nonexisting_model_dir = '/sup23/123s_nets2'


def mocked_requests_get(*args, **kwargs):
    class MockResponse:
        def __init__(self, content, status_code):
            self.content = content
            self.status_code = status_code

        def get(self):
            return self.content

        def raise_for_status(self):
            if self.status_code == 404:
                raise requests.exceptions.HTTPError

    if args[0] == valid_remote_url:
        return MockResponse("Hello world!", 200)

    return MockResponse(None, 404)


def test_get_remote_file_model_dir_doesnt_exist_error():
    with pytest.raises(NotADirectoryError):
        get_remote_file(
            remote_url=valid_remote_url,
            model_dir=nonexisting_model_dir,
        )


@patch("builtins.open", new_callable=mock_open, read_data="data")
@mock.patch('requests.get', side_effect=mocked_requests_get)
def test_get_remote_file(mock_get, mock_file):
    get_remote_file(
        remote_url=valid_remote_url,
        model_dir=existing_model_dir,
        overwrite=False,
    )
    mock_file.assert_called_with(os.path.join(existing_model_dir, 'test.txt'), "wb")


@mock.patch('requests.get', side_effect=mocked_requests_get)
def test_get_remote_file_remote_doesnt_exist(mock_get):
    with pytest.raises(requests.exceptions.HTTPError):
        get_remote_file(
            remote_url=valid_remote_url_file_not_exists,
            model_dir=existing_model_dir,
            overwrite=False,
        )


@mock.patch("subprocess.Popen")
def test_get_cores(mock_subproc_popen):
    with open('tests/outputs/subprocess_lscpu_icx8380.json', 'r') as f:
        content = f.read()
    process_mock = mock.Mock()
    attrs = {"communicate.return_value": (content, None)}
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    # Get all logical cores
    cores = [int(c) for c in get_cores().split(',')]
    assert len(cores) == 160
    assert list(range(160)) == sorted(cores)

    # Get all logical cores from socket #1 only
    cores = [int(c) for c in get_cores(sockets=[0]).split(',')]
    assert len(cores) == 80
    assert (list(range(0, 40)) + list(range(80, 120))) == sorted(cores)

    # Get physical cores only
    cores = [int(c) for c in get_cores(use_ht=False).split(',')]
    assert len(cores) == 80
    assert list(range(0, 80)) == sorted(cores)

    # Get physical cores from socket #1 only
    cores = [int(c) for c in get_cores(use_ht=False, sockets=[1]).split(',')]
    assert len(cores) == 40

    # Get first 20 physical cores from socket #1 only
    cores = [int(c) for c in get_cores(use_ht=False, sockets=[1], num_cores=20).split(',')]
    assert len(cores) == 20


def test_split_list() -> None:
    assert [[1, 3], [2, 4]] == split_list([1, 2, 3, 4], 2)

    assert [[1, 2, 3, 4]] == split_list([1, 2, 3, 4], 1)

    assert [[1], []] == split_list([1], 2)

    assert [[1], [2], [3], [4], []] == split_list([1, 2, 3, 4], 5)

    assert [[1, 4], [2], [3]] == split_list([1, 2, 3, 4], 3)


def test_check_kwargs_deprecated_acc_accuracy_lat_latency() -> None:
    # For Image Classification task the correct metric name is `bleu` instead of `acc` that was used in the past.
    old = {
        'supernet': 'ofa_resnet50',
        'optimization_metrics': ['lat', 'acc', 'macs', 'params'],
        'measurements': ['lat', 'acc', 'macs', 'params'],
        'random_param': ['lat', 'acc', 'macs', 'params'],
    }

    new = {
        'supernet': 'ofa_resnet50',
        'optimization_metrics': ['latency', 'accuracy_top1', 'macs', 'params'],
        'measurements': ['latency', 'accuracy_top1', 'macs', 'params'],
        'random_param': ['lat', 'acc', 'macs', 'params'],
    }
    updated_old = check_kwargs_deprecated(**old)

    assert new == updated_old

    same_1 = {
        'supernet': 'ofa_resnet50',
        'optimization_metrics': ['latency', 'accuracy_top1', 'macs', 'params'],
        'measurements': ['latency', 'accuracy_top1', 'macs', 'params'],
        'random_param': ['lat', 'acc', 'macs', 'params'],
    }

    same_2 = {
        'supernet': 'ofa_resnet50',
        'optimization_metrics': ['latency', 'accuracy_top1', 'macs', 'params'],
        'measurements': ['latency', 'accuracy_top1', 'macs', 'params'],
        'random_param': ['lat', 'acc', 'macs', 'params'],
    }
    same_2_updated = check_kwargs_deprecated(**same_1)

    assert same_2_updated == same_2


def test_check_kwargs_deprecated_acc_bleu() -> None:
    # For Language Translation task the correct metric name is `bleu` instead of `acc` that was used in the past.
    old = {
        'supernet': 'transformer_lt_wmt_en_de',
        'optimization_metrics': ['lat', 'acc', 'macs', 'params'],
        'measurements': ['lat', 'acc', 'macs', 'params'],
        'random_param': ['lat', 'acc', 'macs', 'params'],
    }

    new = {
        'supernet': 'transformer_lt_wmt_en_de',
        'optimization_metrics': ['latency', 'bleu', 'macs', 'params'],
        'measurements': ['latency', 'bleu', 'macs', 'params'],
        'random_param': ['lat', 'acc', 'macs', 'params'],
    }
    updated_old = check_kwargs_deprecated(**old)

    assert new == updated_old



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dynast.utils.cache import Cache


def test_cache_add_new(tmp_path):
    cache = Cache("ut", tmp_path)
    payload = {"a": 1, "b": 2}
    cache.update(key="test1", payload=payload)
    assert cache.get(key="test1") == payload


def test_cache_update_existing(tmp_path):
    cache = Cache("ut", tmp_path)
    payload = {"a": 1, "b": 2}
    cache.update(key="test1", payload=payload)

    updated_payload = {"a": 1, "b": "b", "c": 3}
    cache.update(key="test1", payload=updated_payload)

    assert cache.get(key="test1") == updated_payload


def test_cache_get_missing_key_none(tmp_path):
    cache = Cache("ut", tmp_path)
    payload = {"a": 1, "b": 2}
    cache.update(key="test1", payload=payload)

    assert cache.get(key="test2") == None


def test_key(tmp_path):
    cache = Cache("ut", tmp_path)
    assert cache.key([0, 0, 0, 0, 1, 1]) == "0_0_0_0_1_1"



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import torch
import torchvision
import torchvision.datasets as datasets
import torchvision.transforms as transforms

from dynast.utils import measure_time


def _dataset_fraction(dataset: torchvision.datasets.DatasetFolder, fraction: float):
    # Use random subset of validation data if valid fraction specified
    if (fraction > 0.0) and (fraction < 1.0):
        random_indices = torch.randperm(len(dataset))
        example_count = round(fraction * len(dataset))
        dataset = torch.utils.data.Subset(dataset, random_indices[:example_count])
    return dataset


class Dataset(object):
    PATH = ""

    @staticmethod
    def name():
        raise NotImplementedError()

    @staticmethod
    def get(dataset_name: str) -> "Dataset":
        dataset_name = dataset_name.lower()
        if "imagenet" == dataset_name:
            return ImageNet
        elif "imagenette" == dataset_name:
            return Imagenette
        elif "cifar10" == dataset_name:
            return CIFAR10
        else:
            raise Exception("Not a valid dataset name.")

    @staticmethod
    def train_transforms() -> transforms.Compose:
        raise NotImplementedError()

    @staticmethod
    def val_transforms() -> transforms.Compose:
        raise NotImplementedError()

    @staticmethod
    @measure_time
    def train_dataloader() -> torch.utils.data.DataLoader:
        raise NotImplementedError()

    @staticmethod
    @measure_time
    def validation_dataloader() -> torch.utils.data.DataLoader:
        raise NotImplementedError()

    @staticmethod
    @measure_time
    def test_dataloader() -> torch.utils.data.DataLoader:
        raise NotImplementedError()


class ImageNet(Dataset):
    PATH = "/datasets/imagenet-ilsvrc2012/"

    @staticmethod
    def name() -> str:
        return "ImageNet"

    @staticmethod
    def _transform_normalize() -> transforms.Normalize:
        return transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])

    @staticmethod
    def train_transforms(image_size: int) -> transforms.Compose:
        return transforms.Compose(
            [
                transforms.RandomResizedCrop(image_size),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                ImageNet._transform_normalize(),
            ]
        )

    @staticmethod
    def val_transforms(image_size: int) -> transforms.Compose:
        return transforms.Compose(
            [
                transforms.Resize(int(image_size / 0.875)),
                transforms.CenterCrop(image_size),
                transforms.ToTensor(),
                ImageNet._transform_normalize(),
            ]
        )

    @staticmethod
    @measure_time
    def train_dataloader(
        batch_size: int,
        image_size: int = 224,
        train_dir: str = None,
        shuffle: bool = True,
        num_workers: int = 16,
    ) -> torch.utils.data.DataLoader:
        if not train_dir:
            train_dir = os.path.join(ImageNet.PATH, "train")

        train_dataset = datasets.ImageFolder(
            train_dir,
            ImageNet.train_transforms(image_size),
        )
        train_loader = torch.utils.data.DataLoader(
            train_dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=True if torch.cuda.is_available() else False,
            drop_last=True,
        )

        return train_loader

    @staticmethod
    @measure_time
    def validation_dataloader(
        batch_size: int,
        image_size: int = 224,
        val_dir: str = None,
        shuffle: bool = False,
        num_workers: int = 16,
        fraction: float = 1.0,
    ) -> torch.utils.data.DataLoader:
        if not val_dir:
            val_dir = os.path.join(ImageNet.PATH, "val")

        val_dataset = datasets.ImageFolder(
            val_dir,
            ImageNet.val_transforms(image_size),
        )

        val_dataset = _dataset_fraction(val_dataset, fraction)

        val_sampler = torch.utils.data.SequentialSampler(val_dataset)
        val_loader = torch.utils.data.DataLoader(
            val_dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=True if torch.cuda.is_available() else False,
            sampler=val_sampler,
            drop_last=True,
        )
        return val_loader


class Imagenette(Dataset):
    """Imagenette is a subset of 10 easily classified classes from Imagenet.
    More: https://github.com/fastai/imagenette
    """

    PATH = "/store/nosnap/datasets/imagenette2-320"

    def __init__(self) -> None:
        super().__init__()

    @staticmethod
    def name() -> str:
        return "Imagenette"

    @staticmethod
    @measure_time
    def train_dataloader(
        batch_size: int,
        image_size: int = 224,
        train_dir: str = None,
        shuffle: bool = True,
        num_workers: int = 16,
    ) -> torch.utils.data.DataLoader:
        if not train_dir:
            train_dir = os.path.join(Imagenette.PATH, "train")

        train_dataset = datasets.ImageFolder(
            train_dir,
            ImageNet.train_transforms(image_size),
        )
        train_loader = torch.utils.data.DataLoader(
            train_dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=True if torch.cuda.is_available() else False,
            drop_last=True,
        )

        return train_loader

    @staticmethod
    @measure_time
    def validation_dataloader(
        batch_size: int,
        image_size: int = 224,
        val_dir: str = None,
        shuffle: bool = False,
        num_workers: int = 16,
        fraction: float = 1.0,
    ) -> torch.utils.data.DataLoader:
        if not val_dir:
            val_dir = os.path.join(Imagenette.PATH, "val")

        val_dataset = datasets.ImageFolder(
            val_dir,
            ImageNet.val_transforms(image_size),
        )

        val_dataset = _dataset_fraction(val_dataset, fraction)

        val_sampler = torch.utils.data.SequentialSampler(val_dataset)
        val_loader = torch.utils.data.DataLoader(
            val_dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=True if torch.cuda.is_available() else False,
            sampler=val_sampler,
            drop_last=True,
        )
        return val_loader


class CIFAR10(Dataset):
    PATH = "./cifar10"

    @staticmethod
    def name() -> str:
        return "CIFAR10"

    @staticmethod
    def _transform() -> transforms.Compose:
        return transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
            ]
        )

    @staticmethod
    def train_transforms() -> transforms.Compose:
        return CIFAR10._transform()

    @staticmethod
    def val_transforms() -> transforms.Compose:
        return CIFAR10._transform()

    @staticmethod
    @measure_time
    def train_dataloader(
        batch_size: int,
        shuffle: bool = True,
        num_workers=16,
    ) -> torch.utils.data.DataLoader:
        trainset = torchvision.datasets.CIFAR10(
            root=CIFAR10.PATH,
            train=True,
            download=True,
            transform=CIFAR10.train_transforms(),
        )
        trainloader = torch.utils.data.DataLoader(
            trainset,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=True if torch.cuda.is_available() else False,
        )
        return trainloader

    @staticmethod
    @measure_time
    def validation_dataloader(
        batch_size: int,
        shuffle: bool = False,
        num_workers=16,
        fraction: float = 1.0,
    ) -> torch.utils.data.DataLoader:
        testset = torchvision.datasets.CIFAR10(
            root=CIFAR10.PATH,
            train=False,
            download=True,
            transform=CIFAR10.val_transforms(),
        )

        testset = _dataset_fraction(testset, fraction)

        testloader = torch.utils.data.DataLoader(
            testset,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=True if torch.cuda.is_available() else False,
        )
        return testloader


class CustomDataset(Dataset):
    @staticmethod
    def name() -> str:
        return "CustomDataset"

    @staticmethod
    @measure_time
    def train_dataloader(
        train_dir: str,
        batch_size: int,
        data_transforms: transforms.Compose,
        shuffle: bool = True,
        num_workers: int = 16,
    ) -> torch.utils.data.DataLoader:
        image_dataset = datasets.ImageFolder(train_dir, data_transforms)
        return torch.utils.data.DataLoader(
            image_dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=True if torch.cuda.is_available() else False,
        )

    @staticmethod
    @measure_time
    def validation_dataloader(
        val_dir: str,
        batch_size: int,
        data_transforms: transforms.Compose,
        shuffle: bool = False,
        num_workers: int = 16,
        fraction: float = 1.0,
    ) -> torch.utils.data.DataLoader:
        image_dataset = datasets.ImageFolder(val_dir, data_transforms)

        image_dataset = _dataset_fraction(image_dataset, fraction)

        return torch.utils.data.DataLoader(
            image_dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=True if torch.cuda.is_available() else False,
        )



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import pytest

from tests.checks.helpers import get_python_files

ROOT_DIR = os.getcwd()


def get_license_template():
    """Retrive license from LICENSE.md to be used as a template to compare against."""

    license_path = os.path.join(ROOT_DIR, 'LICENSE.md')
    license_content = ''
    with open(license_path) as f:
        lines = f.readlines()
        for line in lines:
            line = line.strip()
            if line.startswith('##'):
                # This will halt on the 3rd party licenses.
                break

            if line and not line.startswith('#'):
                license_content += line + ' '
    return license_content


def get_license_header(path):
    license_content = ''
    with open(path) as f:
        lines = f.readlines()
        for line in lines:
            line = line.strip()
            if line:
                if line.startswith('#'):
                    # NOTE(macsz): Downside of this approach is that all license header lines have to start with
                    # a '#' comment sign.
                    line = line.replace('#', '').strip()
                    if line:
                        license_content += line + ' '
                else:
                    break
    return license_content


license_target = get_license_template()


@pytest.mark.parametrize('fp', get_python_files(ROOT_DIR))
def test_check_license(fp):
    license_file = get_license_header(fp)
    assert license_target in license_file, f"Wrong license header in `{fp}`"



# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

from dynast.dynast_manager import DyNAS


def _main(args):
    agent = DyNAS(
        supernet=args.supernet,
        optimization_metrics=args.optimization_metrics,
        measurements=args.measurements,
        search_tactic=args.search_tactic,
        num_evals=args.num_evals,
        results_path=args.results_path,
        dataset_path=args.dataset_path,
        seed=args.seed,
        population=args.population,
        batch_size=args.batch_size,
        search_algo=args.search_algo,
        verbose=args.verbose,
        supernet_ckpt_path=args.supernet_ckpt_path,
        device=args.device,
        test_fraction=args.test_fraction,
        dataloader_workers=args.dataloader_workers,
        distributed=args.distributed,
    )

    results = agent.search()

    print('Search results: ', results)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--supernet',
        default='ofa_mbv3_d234_e346_k357_w1.0',
        type=str,
        help='Super-network',
        choices=[
            'ofa_resnet50',
            'ofa_mbv3_d234_e346_k357_w1.0',
            'ofa_mbv3_d234_e346_k357_w1.2',
            'ofa_proxyless_d234_e346_k357_w1.3',
            'transformer_lt_wmt_en_de',
            'bert_base_sst2',
        ],
    )
    parser.add_argument('--seed', default=42, type=int, help='Random Seed')
    # TODO(macsz) might be better to list allowed elements here (or in help message)
    parser.add_argument(
        '--optimization_metrics',
        default=['accuracy_top1', 'macs'],  # TODO*macsz) Consider just 'accuracy',
        # as it translates better for both image classification
        # and Transformer LT
        nargs='*',
        type=str,
        help='Metrics that will be optimized for during search.',
    )
    parser.add_argument(
        '--measurements',
        default=['accuracy_top1', 'macs', 'params', 'latency'],  # TODO(macsz) Ditto
        nargs='*',
        type=str,
        help='Measurements during search.',
    )
    parser.add_argument('-d', '--device', default='cpu', type=str, help='Target device to run measurements on.')
    parser.add_argument('--num_evals', default=250, type=int, help='Total number of evaluations during search.')
    parser.add_argument('--batch_size', default=128, type=int, help='Batch size for latency measurement calculation.')
    parser.add_argument(
        '--test_fraction',
        default=1.0,
        type=float,
        help='Fraction of the validation data to be used for testing and evaluation.',
    )
    parser.add_argument('--dataloader_workers', default=4, type=int, help='How many workers to use when loading data.')
    parser.add_argument('--population', default=50, type=int, help='Population size for each generation')
    parser.add_argument('--results_path', required=True, type=str, help='Path to store search results, csv format')
    parser.add_argument('--dataset_path', default='/datasets/imagenet-ilsvrc2012', type=str, help='')
    parser.add_argument('--supernet_ckpt_path', default=None, type=str, help='Path to supernet checkpoint.')

    parser.add_argument(
        '--search_tactic',
        default='linas',
        choices=['linas', 'evolutionary', 'random'],
        help='Search tactic, currently supports: ["linas", "evolutionary" , "random"]',
    )
    parser.add_argument(
        '--search_algo',
        default='nsga2',
        choices=['cmaes', 'nsga2', 'age', 'unsga3', 'ga', 'ctaea', 'moead'],
        help='Search algorithm, currently supports: ["cmaes", "nsga2", "age", "unsga3", "ga", "ctaea", "moead"]',
    )
    parser.add_argument('-v', '--verbose', action='store_true', help='Print more information.')

    dist_parser = parser.add_argument_group('Distributed search options')
    parser.add_argument(
        '--distributed',
        action='store_true',
        help='If set, a distributed implementation of the search algorithm will be used.',
    )
    dist_parser.add_argument(
        "--local_rank", type=int, help="Local rank. Necessary for using the torch.distributed.launch utility."
    )
    dist_parser.add_argument("--backend", type=str, default="gloo", choices=['gloo'])

    args = parser.parse_args()

    _main(args)


if __name__ == '__main__':
    main()
